from typing import Optional, List
from collections import UserDict
try:
    import orjson as json
except ImportError:
    try:
        import rapidjson as json
    except ImportError:
        try:
            import simplejson as json
        except ImportError:
            import json

from validator_collection import validators, errors as validator_errors


from highcharts_core.decorators import validate_types, class_sensitive
from highcharts_core.metaclasses import HighchartsMeta, JavaScriptDict


class AttributeObject(JavaScriptDict):
    """Special :class:`dict <python:dict>` class used to construct an arbitrary set of
    SVG attributes. Each key represents the key to the attribute (expressed as a string),
    while the value represents the value of the attribute.

    Keys are validated to be valid variable names.
    """

    def __setitem__(self, key, item):
        validate_key = False
        try:
            validate_key = key not in self
        except AttributeError:
            validate_key = True

        if validate_key:
            try:
                key = validators.variable_name(key, allow_empty = False)
            except validator_errors.InvalidVariableNameError as error:
                if '-' in key:
                    try:
                        test_key = key.replace('-', '_')
                        validators.variable_name(test_key, allow_empty = False)
                    except validator_errors.InvalidVariableNameError:
                        raise error
                else:
                    raise error

        super().__setitem__(key, item)


class ASTNode(HighchartsMeta):
    """Generic representation of an AST node."""

    def __init__(self, **kwargs):
        self._attributes = None
        self._children = []
        self._tag_name = None
        self._text_content = None

        self.attributes = kwargs.get('attributes', None)
        self.children = kwargs.get('children', [])
        self.tag_name = kwargs.get('tag_name', None)
        self.text_content = kwargs.get('text_content', None)

    @property
    def attributes(self) -> Optional[AttributeObject]:
        """Map of attributes that are used to describe the node.

        Basically corresponds to a :class:`dict <python:dict>`, where keys have to
        generally conform to standard variable-naming conventions and values are basic
        literals (string, array, numer, boolean, :class:`dict <python:dict>`, etc.)

        :rtype: :class:`SVGAttributeMap` or :obj:`None <python:None>`
        """
        return self._attributes

    @attributes.setter
    @class_sensitive(AttributeObject)
    def attributes(self, value):
        self._attributes = value

    @property
    def children(self) -> List:
        """Collection of :class:`ASTNode` instances that should be considered children
        of this node.

        :rtype: :class:`list <python:list>` of :class:`ASTNode` objects
        """
        return self._children

    @children.setter
    def children(self, value):
        if not value:
            self._children = []
        else:
            value = validate_types(value,
                                   types = self.__class__,
                                   allow_none = False,
                                   force_iterable = True)
            self._children = value

    @property
    def tag_name(self) -> Optional[str]:
        """The name given to the SVG tag.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._tag_name

    @tag_name.setter
    def tag_name(self, value):
        self._tag_name = validators.string(value, allow_empty = True)

    @property
    def text_content(self) -> Optional[str]:
        """Text content that populates the tag, if applicable.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._text_content

    @text_content.setter
    def text_content(self, value):
        self._text_content = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'attributes': as_dict.get('attributes', None),
            'children': as_dict.get('children', []),
            'tag_name': as_dict.get('tagName', None),
            'text_content': as_dict.get('textContent', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'attributes': self.attributes,
            'children': self.children,
            'tagName': self.tag_name,
            'textContent': self.text_content
        }

        return untrimmed


class ASTMap(JavaScriptDict):
    """Special :class:`dict <python:dict>` class used to construct a JavaScript object
    literal map. Each key represents a key or node within that map, while
    the value for that key is an :class:`ASTNode`.

    Keys are validated to be valid variable names, while each value is validated to be an
    :class:`ASTNode`.

    """
    _valid_value_types = ASTNode
    _allow_empty_value = True


class TextPath(HighchartsMeta):
    """Options for a label text which should follow marker's shape.

    .. note::

      Border and background are disabled for a label that follows a path.

    .. warning::

      Only SVG-based renderer supports this option. Setting :meth:`DataLabel.use_html` to
      ``True`` will disable this option.

    """

    def __init__(self, **kwargs):
        self._attributes = None
        self._enabled = None

        self.attributes = kwargs.get('attributes', None)
        self.enabled = kwargs.get('enabled', None)

    @property
    def attributes(self) -> Optional[AttributeObject]:
        """Presentation attributes for the text path.

        :rtype: :class:`AttributeObject` or :obj:`None <python:None>`
        """
        return self._attributes

    @attributes.setter
    @class_sensitive(AttributeObject)
    def attributes(self, value):
        self._attributes = value

    @property
    def enabled(self) -> Optional[bool]:
        """If ``True``, enables the use of a text path for links or marker data labels.
        Defaults to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'attributes': as_dict.get('attributes', None),
            'enabled': as_dict.get('enabled', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'attributes': self.attributes,
            'enabled': self.enabled
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.markers import Marker


class Zone(HighchartsMeta):
    """A zone defined within a series.

    Zones can be applied to the X axis, Y axis or Z axis for bubbles, according to the
    :meth:`zone_axis <AreaOptions.zone_axis>` option.

    """

    def __init__(self, **kwargs):
        self._class_name = None
        self._color = None
        self._dash_style = None
        self._fill_color = None
        self._value = None

        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.dash_style = kwargs.get('dash_style', None)
        self.fill_color = kwargs.get('fill_color', None)
        self.value = kwargs.get('value', None)

    @property
    def class_name(self) -> Optional[str]:
        """A custom class name for the zone.

        .. warning::

          Supported in :term:`styled mode` only.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str | Gradient | Pattern]:
        """Defines the color of the series. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        from highcharts_core import utility_functions
        self._color = utility_functions.validate_color(value)

    @property
    def dash_style(self) -> Optional[str]:
        """Name of the dash style to use for the graph, or for some series types the
        outline of each shape. Defaults to :obj:`None <python:None>`.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._dash_style

    @dash_style.setter
    def dash_style(self, value):
        if not value:
            self._dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'dash_style expects a recognized value'
                                                  f', but received: {value}')
            self._dash_style = value

    @property
    def fill_color(self) -> Optional[str | Gradient | Pattern]:
        """Fill color or gradient for the area.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`
        """
        return self._fill_color

    @fill_color.setter
    def fill_color(self, value):
        from highcharts_core import utility_functions
        self._fill_color = utility_functions.validate_color(value)

    @property
    def value(self) -> Optional[int | float | Decimal]:
        """The value up to where the zone extends. If :obj:`None <python:None>`, the zone
        stretches to the last value in the series. Defaults to :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._value

    @value.setter
    def value(self, value_):
        self._value = validators.numeric(value_, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'dash_style': as_dict.get('dashStyle', None),
            'fill_color': as_dict.get('fillColor', None),
            'value': as_dict.get('value', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'className': self.class_name,
            'color': self.color,
            'dashStyle': self.dash_style,
            'fillColor': self.fill_color,
            'value': self.value
        }

        return untrimmed


class ClusterZone(HighchartsMeta):
    """A zone defined for a group of Clusters."""

    def __init__(self, **kwargs):
        self._class_name = None
        self._from = None
        self._marker = None
        self._to = None

        self.class_name = kwargs.get('class_name', None)
        self.from_ = kwargs.get('from_', None)
        self.marker = kwargs.get('marker', None)
        self.to = kwargs.get('to', None)

    @property
    def class_name(self) -> Optional[str]:
        """A custom class name for the zone.

        .. warning::

          Supported in :term:`styled mode` only.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def from_(self) -> Optional[int | float | Decimal]:
        """The value where the zone starts.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._from

    @from_.setter
    def from_(self, value):
        self._from = validators.numeric(value, allow_empty = True)

    @property
    def marker(self) -> Optional[Marker]:
        """Settings for the cluster marker belonging to the zone.

        :rtype: :class:`Marker` or :obj:`None <python:None>`
        """
        return self._marker

    @marker.setter
    @class_sensitive(Marker)
    def marker(self, value):
        self._marker = value

    @property
    def to(self) -> Optional[int | float | Decimal]:
        """The value where the zone ends.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._to

    @to.setter
    def to(self, value):
        self._to = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'class_name': as_dict.get('className', None),
            'from_': as_dict.get('from', None),
            'marker': as_dict.get('marker', None),
            'to': as_dict.get('to', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'className': self.class_name,
            'from': self.from_,
            'marker': self.marker,
            'to': self.to
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.options.series.area import AreaSeries
from highcharts_core.options.plot_options.bellcurve import BellCurveOptions
from highcharts_core.utility_functions import mro__to_untrimmed_dict


class BellCurveSeries(AreaSeries, BellCurveOptions):
    """Options to configure a Bell Curve series.

    A bell curve is an areaspline series which represents the probability density
    function of the normal distribution. It calculates mean and standard deviation of
    the base series data and plots the curve according to the calculated parameters.

    .. figure:: ../../../_static/bellcurve-example.png
      :alt: Bell Curve Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._base_series = None

        self.base_series = kwargs.get('base_series', None)

        super().__init__(**kwargs)

    @property
    def data(self) -> None:
        """The collection of data points for the series. Defaults to
        :obj:`None <python:None>`.

        .. warning::

          All Bell Curve Series by definition return :obj:`None <python:None>` for their
          data. They are a special series, drawn in relationship to the
          :meth:`base_series <BellCurveSeries.base_series>` specified, and do not receive
          independent data points of their own.

        :rtype: :obj:`None <python:None>`
        """
        return None

    @data.setter
    def data(self, value):
        pass

    @property
    def base_series(self) -> Optional[int | str]:
        """An integer identifying the index to use for the base series, or a string
        representing the id of the series. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`int <python:int>` or :class:`str <python:str>` or
          :obj:`None <python:None>`
        """
        return self._base_series

    @base_series.setter
    def base_series(self, value):
        if value is None:
            self._base_series = None
        else:
            try:
                value = validators.string(value)
            except (TypeError, ValueError):
                value = validators.integer(value)

            self._base_series = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'fill_color': as_dict.get('fillColor', None),
            'fill_opacity': as_dict.get('fillOpacity', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'negative_fill_color': as_dict.get('negativeFillColor', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'track_by_area': as_dict.get('trackByArea', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'data': as_dict.get('data', None),
            'id': as_dict.get('id', None),
            'index': as_dict.get('index', None),
            'legend_index': as_dict.get('legendIndex', None),
            'name': as_dict.get('name', None),
            'stack': as_dict.get('stack', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),
            'z_index': as_dict.get('zIndex', None),

            'intervals': as_dict.get('intervals', None),

            'base_series': as_dict.get('baseSeries', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'baseSeries': self.base_series,
            'data': self.data
        }

        parents_as_dict = mro__to_untrimmed_dict(self, in_cls = in_cls) or {}
        for key in parents_as_dict:
            untrimmed[key] = parents_as_dict[key]

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError

from dotenv import load_dotenv

from highcharts_core.options.data import Data as cls
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal, run_pyspark_tests

STANDARD_PARAMS = [
    ({}, None),
    ({
      'csv_url': 'https://www.somewhere.dev/test.csv',
      'data_refresh_rate': 3,
      'decimal_point': '.',
      'enable_polling': True,
      'first_row_as_names': True,
      'item_delimiter': ','
    }, None),
    ({
      'csv_url': '/test.csv',
      'data_refresh_rate': 3,
      'decimal_point': '.',
      'enable_polling': True,
      'first_row_as_names': True,
      'item_delimiter': ','
    }, None),

    ({
      'csv_url': 123,
      'data_refresh_rate': 3,
      'decimal_point': '.',
      'enable_polling': True,
      'first_row_as_names': True,
      'item_delimiter': ','
    }, (ValueError, TypeError)),
]

load_dotenv()


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('data/01.js', False, None),
    ('data/02.js', False, None),

    ('data/error-01.js', False, (errors.HighchartsValueError,
                                 errors.HighchartsParseError,
                                 JSONDecodeError,
                                 ValueError)),
    ('data/error-02.js', False, (errors.HighchartsValueError,
                                 errors.HighchartsParseError,
                                 JSONDecodeError,
                                 ValueError)),

    ('data/01.js', True, None),
    ('data/02.js', True, None),

    ('data/error-01.js', True, (errors.HighchartsValueError,
                                errors.HighchartsParseError,
                                JSONDecodeError,
                                ValueError)),
    ('data/error-02.js', True, (errors.HighchartsValueError,
                                errors.HighchartsParseError,
                                JSONDecodeError,
                                ValueError)),
])
def test_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


@pytest.mark.parametrize('filename, kwargs, error', [
    ('test-data-files/nst-est2019-01-transposed.csv', {}, None),
    ('test-data-files/nst-est2019-01-transposed.csv',
     {'represent_as': 'html'},
     None),

])
def test_from_pandas(input_files, filename, kwargs, error):
    import pandas

    input_file = check_input_file(input_files, filename)
    df = pandas.read_csv(input_file)

    if not error:
        result = cls.from_pandas(df, **kwargs)
        assert result is not None
        assert isinstance(result, cls) is True

        if kwargs.get('represent_as', 'csv') == 'csv':
            assert result.csv is not None
            assert result.table is None
            assert isinstance(result.csv, str) is True
        else:
            assert result.csv is None
            assert result.table is not None
            assert isinstance(result.table, str) is True
    else:
        with pytest.raises(error):
            result = cls.from_pandas(df, **kwargs)


@pytest.mark.parametrize('filename, kwargs, error', [
    ('test-data-files/nst-est2019-01-transposed.csv', {}, None),
    ('test-data-files/nst-est2019-01-transposed.csv',
     {'consolidation': 'repartition'},
     None),
    ('test-data-files/nst-est2019-01-transposed.csv',
     {'consolidation': 'coalesce'},
     None),
    ('test-data-files/nst-est2019-01-transposed.csv',
     {'consolidation': None}, 
     None),

])
def test_from_pyspark(input_files, run_pyspark_tests, filename, kwargs, error):
    if run_pyspark_tests:
        from pyspark.sql import SparkSession

        spark = SparkSession.builder.appName('highcharts.tests').getOrCreate()

        input_file = check_input_file(input_files, filename)
        df = spark.read.csv(input_file)

        if not error:
            result = cls.from_pyspark(df, **kwargs)
            assert result is not None
            assert isinstance(result, cls) is True

            assert result.csv is not None
            assert result.table is None
            assert isinstance(result.csv, str) is True
        else:
            with pytest.raises(error):
                result = cls.from_pyspark(df, **kwargs)



"""Tests for ``highcharts.no_data``."""

import pytest
from copy import deepcopy

from json.decoder import JSONDecodeError

from highcharts_core.utility_classes.buttons import ButtonConfiguration as cls
from highcharts_core.utility_classes.buttons import ContextButtonConfiguration as cls2
from highcharts_core.utility_classes.buttons import ExportingButtons as cls3
from highcharts_core import errors, constants
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal, does_kwarg_value_match_result

STANDARD_PARAMS = [
    ({}, None),
    ({
      'enabled': True,
      'text': 'Button Label',
      'theme': {
          'fill': '#fff',
          'stroke': '#ccc'
      },
      'y': 0
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_ButtonConfiguration__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_ButtonConfiguration__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_ButtonConfiguration_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_ButtonConfiguration_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/buttons/01.js', False, None),

    ('utility_classes/buttons/error-01.js', False, (errors.HighchartsValueError,
                                                    errors.HighchartsParseError,
                                                    JSONDecodeError,
                                                    TypeError,
                                                    ValueError)),

    ('utility_classes/buttons/01.js', True, None),

    ('utility_classes/buttons/error-01.js', True, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
])
def test_ButtonConfiguration_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'class_name': 'some-class-name',
      'enabled': True,
      'menu_class_name': 'menu-class',
      'menu_items': ['item1', 'item2'],
      'onclick': """function (event) { return true; }""",
      'symbol': 'menu',
      'symbol_fill': '#ccc',
      'text': 'Button Label',
      'theme': {
          'fill': '#fff',
          'stroke': '#ccc'
      },
      'title_key': 'somevalue',
      'x': 10,
      'y': 0
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_ContextButtonConfiguration__init__(kwargs, error):
    Class__init__(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_ContextButtonConfiguration__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_ContextButtonConfiguration_from_dict(kwargs, error):
    Class_from_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_ContextButtonConfiguration_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/buttons/02.js', False, None),

    ('utility_classes/buttons/error-03.js', False, (errors.HighchartsValueError,
                                                    errors.HighchartsParseError,
                                                    JSONDecodeError,
                                                    TypeError,
                                                    ValueError)),

    ('utility_classes/buttons/02.js', True, None),

    ('utility_classes/buttons/error-03.js', True, (errors.HighchartsValueError,
                                                   errors.HighchartsParseError,
                                                   JSONDecodeError,
                                                   TypeError,
                                                   ValueError)),
])
def test_ContextButtonConfiguration_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls2, input_files, filename, as_file, error)


STANDARD_PARAMS_3 = [
    ({}, None),
    ({
      'button1': {
        'enabled': True,
        'text': 'Button Label',
        'theme': {
            'fill': '#fff',
            'stroke': '#ccc'
        },
        'y': 0
      },
      'button2': {
        'className': 'some-class-name',
        'enabled': True,
        'menuClassName': 'menu-class',
        'menuItems': ['item1', 'item2'],
        'onclick': """function (event) { return true; }""",
        'symbol': 'menu',
        'symbolFill': '#ccc',
        'text': 'Button Label',
        'theme': {
            'fill': '#fff',
            'stroke': '#ccc'
        },
        'titleKey': 'somevalue',
        'x': 10,
        'y': 0
      }
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_ExportingButtons__init__(kwargs, error):
    kwargs_copy = {}
    for key in kwargs:
        if isinstance(kwargs[key], dict):
            kwargs_copy[key] = deepcopy(kwargs[key])
        kwargs_copy[key] = kwargs[key]
    print('before init')
    print(kwargs_copy)
    if 'item1' in kwargs_copy:
        print(id(kwargs_copy['item1']))
        print(id(kwargs['item1']))
    if not error:
        result = cls3(**kwargs)
        assert result is not None
        assert isinstance(result, cls3) is True
        for key in kwargs:
            if key in ['context_button', 'contextButton']:
                continue
            if isinstance(kwargs[key], str) and kwargs[key].startswith('function'):
                continue
            if isinstance(kwargs[key], str) and kwargs[key].startswith('class'):
                continue

            print(f'after __init__ for key: {key}')
            print(kwargs_copy[key])
            print(result)
            print(result.get(key))
            assert does_kwarg_value_match_result(kwargs_copy[key],
                                                 result.get(key)) is True
    else:
        with pytest.raises(error):
            result = cls3(**kwargs)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_3)
def test_ExportingButtons__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls3, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_ExportingButtons_from_dict(kwargs, error):
    as_dict = to_js_dict(deepcopy(kwargs))

    if not error:
        instance = cls3.from_dict(as_dict)
        assert instance is not None
        assert isinstance(instance, cls3) is True
        for key in kwargs:
            assert does_kwarg_value_match_result(kwargs[key],
                                                 instance.get(key)) is True
    else:
        with pytest.raises(error):
            instance = cls3.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_3)
def test_ExportingButtons_to_dict(kwargs, error):
    Class_to_dict(cls3, kwargs, error)



from typing import Optional

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.metaclasses import HighchartsMeta


class TableLanguageOptions(HighchartsMeta):
    """Accessibility language options for the data table."""

    def __init__(self, **kwargs):
        self._table_summary = None
        self._view_as_data_table_button_text = None

        self.table_summary = kwargs.get('table_summary', None)
        self.view_as_data_table_button_text = kwargs.get('view_as_data_table_button_text',
                                                         None)

    @property
    def table_summary(self) -> Optional[str]:
        """Defaults to
        ``'Table representation of chart.'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._table_summary

    @table_summary.setter
    def table_summary(self, value):
        self._table_summary = validators.string(value, allow_empty = True)

    @property
    def view_as_data_table_button_text(self) -> Optional[str]:
        """Defaults to
        ``'View as data table, {chartTitle}'``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._view_as_data_table_button_text

    @view_as_data_table_button_text.setter
    def view_as_data_table_button_text(self, value):
        self._view_as_data_table_button_text = validators.string(value,
                                                                 allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'table_summary': as_dict.get('tableSummary', None),
            'view_as_data_table_button_text': as_dict.get('viewAsDataTableButtonText',
                                                          None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'tableSummary': self.table_summary,
            'viewAsDataTableButtonText': self.view_as_data_table_button_text
        }

        return untrimmed



"""Tests for ``highcharts.no_data``."""

import pytest

from json.decoder import JSONDecodeError
from copy import deepcopy
from collections import UserDict

from highcharts_core.utility_classes.menus import MenuItem as cls
from highcharts_core.utility_classes.menus import MenuObject as cls2
from highcharts_core import errors
from tests.fixtures import input_files, check_input_file, to_camelCase, to_js_dict, \
    Class__init__, Class__to_untrimmed_dict, Class_from_dict, Class_to_dict, \
    Class_from_js_literal, does_kwarg_value_match_result

STANDARD_PARAMS = [
    ({}, None),
    ({
      'onclick': """function() { return true; }""",
      'text': 'My Menu Item',
      'text_key': 'my-menu-item',
      'separator': False
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_MenuItem__init__(kwargs, error):
    Class__init__(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS)
def test_MenuItem__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_MenuItem_from_dict(kwargs, error):
    Class_from_dict(cls, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS)
def test_MenuItem_to_dict(kwargs, error):
    Class_to_dict(cls, kwargs, error)


@pytest.mark.parametrize('filename, as_file, error', [
    ('utility_classes/menus/01.js', False, None),

    ('utility_classes/menus/error-01.js', False, (errors.HighchartsValueError,
                                                  errors.HighchartsParseError,
                                                  JSONDecodeError,
                                                  TypeError,
                                                  ValueError)),

    ('utility_classes/menus/01.js', True, None),

    ('utility_classes/menus/error-01.js', True, (errors.HighchartsValueError,
                                                 errors.HighchartsParseError,
                                                 JSONDecodeError,
                                                 TypeError,
                                                 ValueError)),
])
def test_MenuItem_from_js_literal(input_files, filename, as_file, error):
    Class_from_js_literal(cls, input_files, filename, as_file, error)


STANDARD_PARAMS_2 = [
    ({}, None),
    ({
      'item1': cls(text = 'My Menu Item',
                   text_key = 'item1',
                   separator = False),
      'item2': cls(text = 'Second Menu Item',
                   text_key = 'item2',
                   separator = False)
    }, None),
    ({
     'item1': {
         'onclick': """function() { return true; }""",
         'text': 'My Menu Item',
         'textKey': 'my-menu-item',
         'separator': False
     },
     'item2': {
         'onclick': """function() { return true; }""",
         'text': 'Second Menu Item',
         'textKey': 'my-menu-item2',
         'separator': False
     },
    }, None),
]


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_MenuObject__init__(kwargs, error):
    kwargs_copy = {}
    for key in kwargs:
        if isinstance(kwargs[key], dict):
            kwargs_copy[key] = deepcopy(kwargs[key])
        kwargs_copy[key] = kwargs[key]
    print('before init')
    print(kwargs_copy)
    if 'item1' in kwargs_copy:
        print(id(kwargs_copy['item1']))
        print(id(kwargs['item1']))
    if not error:
        result = cls2(**kwargs)
        assert result is not None
        assert isinstance(result, cls2) is True
        for key in kwargs:
            if isinstance(kwargs[key], str) and kwargs[key].startswith('function'):
                continue
            if isinstance(kwargs[key], str) and kwargs[key].startswith('class'):
                continue

            print(f'after __init__ for key: {key}')
            print(kwargs_copy[key])
            print(result.get(key))
            assert does_kwarg_value_match_result(kwargs_copy[key],
                                                 result.get(key)) is True
    else:
        with pytest.raises(error):
            result = cls2(**kwargs)


@pytest.mark.parametrize('kwargs, error', STANDARD_PARAMS_2)
def test_MenuObject__to_untrimmed_dict(kwargs, error):
    Class__to_untrimmed_dict(cls2, kwargs, error)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_MenuObject_from_dict(kwargs, error):
    as_dict = to_js_dict(deepcopy(kwargs))

    if not error:
        instance = cls2.from_dict(as_dict)
        assert instance is not None
        assert isinstance(instance, cls2) is True
        for key in kwargs:
            assert does_kwarg_value_match_result(kwargs[key],
                                                 instance.get(key)) is True
    else:
        with pytest.raises(error):
            instance = cls2.from_dict(as_dict)


@pytest.mark.parametrize('kwargs, error',  STANDARD_PARAMS_2)
def test_MenuObject_to_dict(kwargs, error):
    Class_to_dict(cls2, kwargs, error)



"""Implements the :class:`HighchartsOptions` class."""
from typing import Optional, List

from validator_collection import validators

from highcharts_core import utility_functions
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.decorators import class_sensitive
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.options.series.series_generator import create_series_obj

from highcharts_core.options.accessibility import Accessibility
from highcharts_core.options.annotations import Annotation
from highcharts_core.options.boost import Boost
from highcharts_core.options.caption import Caption
from highcharts_core.options.chart import ChartOptions
from highcharts_core.options.axes.color_axis import ColorAxis
from highcharts_core.options.credits import Credits
from highcharts_core.options.data import Data
from highcharts_core.options.defs import MarkerDefinition
from highcharts_core.options.drilldown import Drilldown
from highcharts_core.options.exporting import Exporting
from highcharts_core.global_options.language import Language
from highcharts_core.options.legend import Legend
from highcharts_core.options.loading import Loading
from highcharts_core.options.navigation import Navigation
from highcharts_core.options.no_data import NoData
from highcharts_core.options.pane import Pane
from highcharts_core.options.plot_options import PlotOptions
from highcharts_core.options.plot_options.generic import GenericTypeOptions
from highcharts_core.options.responsive import Responsive
from highcharts_core.options.sonification import SonificationOptions
from highcharts_core.options.subtitle import Subtitle
from highcharts_core.options.time import Time
from highcharts_core.options.title import Title
from highcharts_core.options.tooltips import Tooltip
from highcharts_core.options.axes.x_axis import XAxis
from highcharts_core.options.axes.y_axis import YAxis
from highcharts_core.options.axes.z_axis import ZAxis


class Options(HighchartsMeta):
    """Metaclass which establishes properties shared across different variations of the
    Highcharts configuration objects."""

    def __init__(self, **kwargs):
        self._accessibility = None
        self._annotations = None
        self._caption = None
        self._chart = None
        self._color_axis = None
        self._colors = None
        self._credits = None
        self._data = None
        self._defs = None
        self._exporting = None
        self._language = None
        self._legend = None
        self._loading = None
        self._navigation = None
        self._plot_options = None
        self._responsive = None
        self._series = None
        self._sonification = None
        self._subtitle = None
        self._time = None
        self._title = None
        self._tooltip = None
        self._x_axis = None
        self._y_axis = None

        self.accessibility = kwargs.get('accessibility', None)
        self.annotations = kwargs.get('annotations', None)
        self.caption = kwargs.get('caption', None)
        self.chart = kwargs.get('chart', None)
        self.color_axis = kwargs.get('color_axis', None)
        self.colors = kwargs.get('colors', None)
        self.credits = kwargs.get('credits', None)
        self.data = kwargs.get('data', None)
        self.defs = kwargs.get('defs', None)
        self.exporting = kwargs.get('exporting', None)
        self.language = kwargs.get('language', None)
        self.legend = kwargs.get('legend', None)
        self.loading = kwargs.get('loading', None)
        self.navigation = kwargs.get('navigation', None)
        self.plot_options = kwargs.get('plot_options', None)
        self.responsive = kwargs.get('responsive', None)
        self.sonification = kwargs.get('sonification', None)
        self.subtitle = kwargs.get('subtitle', None)
        self.time = kwargs.get('time', None)
        self.title = kwargs.get('title', None)
        self.tooltip = kwargs.get('tooltip', None)
        self.x_axis = kwargs.get('x_axis', None)
        self.y_axis = kwargs.get('y_axis', None)

        self.series = kwargs.get('series', None)

    @property
    def accessibility(self) -> Optional[Accessibility]:
        """Options for configuring accessibility for the chart.

        .. note::

          Requires the accessibility module to be loaded in the browser. For a description
          of the module and information on its features, see
          `Highcharts Accessibility <https://www.highcharts.com/docs/accessibility/accessibility-module>`_.

        :returns: The accessibility configuration for the chart or :obj:`None <python:None>`
        :rtype: :class:`Accessibility <highcharts.accessibility.Accessibility>` /
          :obj:`None <python:None>`

        :raise HighchartsError: if setting the value to an incompatible type

        """
        return self._accessibility

    @accessibility.setter
    @class_sensitive(Accessibility)
    def accessibility(self, value: Optional[Accessibility | dict | str]):
        self._accessibility = value

    @property
    def annotations(self) -> Optional[List[Annotation]]:
        """A basic type of an annotation. It allows adding custom labels or shapes. The
        items can be tied to points, axis coordinates or chart pixel coordinates.

        :returns: A collection of the annotations applied to the chart.
        :rtype: :class:`list <python:list>` of :class:`Annotation` objects or
          :obj:`None <python:None>` if no annotations are applied

        """
        return self._annotations

    @annotations.setter
    @class_sensitive(Annotation, force_iterable = True)
    def annotations(self, value):
        self._annotations = value

    @property
    def caption(self) -> Optional[Caption]:
        """The chart's caption, which will render below the chart and will be part of
        exported charts.

        .. note::

          The caption can be updated after chart initialization through the
          ``Chart.update`` or ``Chart.caption.update`` JavaScript methods.

        :returns: The chart's caption or :obj:`None <python:None>`
        :rtype: :class:`Caption` / :obj:`None <python:None>`
        """
        return self._caption

    @caption.setter
    @class_sensitive(Caption)
    def caption(self, value):
        self._caption = value

    @property
    def chart(self) -> Optional[ChartOptions]:
        """General options for the chart.

        .. note::

          This property is perhaps one of the most important properties you will use when
          configuring your Highcharts data visualization.

        :returns: A :class:`ChartOptions` configuration object or
          :obj:`None <python:None>`
        :rtype: :class:`ChartOptions` or :obj:`None <python:None>`
        """
        return self._chart

    @chart.setter
    @class_sensitive(ChartOptions)
    def chart(self, value):
        self._chart = value

    @property
    def color_axis(self) -> Optional[List[ColorAxis]]:
        """A color axis for series.

        Visually, the color axis will appear as a gradient or as separate items inside the
        legend, depending on whether the axis is scalar or based on data classes.

        A scalar color axis is represented by a gradient. The colors either range between
        the ``minimum_color`` and the ``maximum_color``, or for more fine grained control
        the colors can be defined in :ref:`stops <ColorAxis.stops>`. Often times, the
        color axis needs to be  adjusted to get the right color spread for the data. In
        addition to stops, consider using a logarithmic axis type, or setting min and max
        to avoid the colors being determined by outliers.

        For supported color formats, please see
        `the documentation article about colors <https://www.highcharts.com/docs/chart-design-and-style/colors>`_.


        When :ref:`data_classes <ColorAxis.data_classes>` are used, the ranges are
        subdivided into separate classes like categories based on their values. This can
        be used for ranges between two values, but also for a true category. However, when
        your data is categorized, it may be as convenient to add each category to a
        separate series.

        .. warning::

          Color axis does not work with: ``sankey``, ``sunburst``, ``dependencywheel``,
          ``networkgraph``, ``wordcloud``, ``venn``, ``gauge`` and ``solidgauge`` series types.

        See the :ref:`Axis` object for programmatic access to the axis.

        :returns: A collection of :class:`ColorAxis` objects defining the color axis to
          apply, or :obj:`None <python:None>`.
        :rtype: :class:`list <python:list>` of :class:`ColorAxis` or
          :obj:`None <python:None>`
        """
        return self._color_axis

    @color_axis.setter
    @class_sensitive(ColorAxis, force_iterable = True)
    def color_axis(self, value):
        self._color_axis = value

    @property
    def colors(self) -> Optional[List[str | Gradient | Pattern]]:
        """An array containing the default colors for the chart's series.

        When all colors are used, new colors are pulled from the start again.

        Default colors can also be set on a series or ``series.type`` basis, see
        :ref:`Column.colors` and :ref:`Pie.colors`.

        .. warning::

          In styled mode, the colors option does not exist.

          Instead, colors are defined in CSS and applied either through series or point
          class names, or through the :ref:`Chart.color_count` option.

          Defaults to:

          .. code-block:: python

            [
                "#7cb5ec",
                "#434348",
                "#90ed7d",
                "#f7a35c",
                "#8085e9",
                "#f15c80",
                "#e4d354",
                "#2b908f",
                "#f45b5b",
                "#91e8e1"
            ]

        :returns: A collection of hex color strings.
        :rtype: :class:`list <python:list>` of :class:`str <python:str>`
        """
        return self._colors

    @colors.setter
    def colors(self, value):
        if not value:
            self._colors = None
        else:
            value = validators.iterable(value)
            self._colors = [utility_functions.validate_color(x) for x in value]

    @property
    def credits(self) -> Optional[Credits]:
        """Highchart by default puts a credits label in the lower right corner of the
        chart. This can be changed using these options.

        :returns: :class:`Credits` configuration or :obj:`None <python:None>`
        :rtype: :class:`Credits` or :obj:`None <python:None>`
        """
        return self._credits

    @credits.setter
    @class_sensitive(Credits)
    def credits(self, value):
        self._credits = value

    @property
    def data(self) -> Optional[Data]:
        """The ``data`` property provides a simplified interface for adding data to a
        chart from sources like CVS, HTML tables, or grid views. See also
        `the tutorial article on the Data module <https://www.highcharts.com/docs/working-with-data/data-module>`_.

        .. warning::

          It requires the ``modules/data.js`` file to be loaded in the browser / client.

        .. warning::

          Please note that the default way of adding data in Highcharts, without the need
          of a module, is through the ``series.type.data`` property.

        :returns: The :class:`Data` object or :obj:`None <python:None>`
        :rtype: :class:`Data` object or :obj:`None <python:None>`
        """
        return self._data

    @data.setter
    @class_sensitive(Data)
    def data(self, value):
        self._data = value

    @property
    def defs(self) -> Optional[MarkerDefinition]:
        """Options for configuring markers for annotations.

        :returns: A :class:`MarkerDefinition` object or
          :obj:`None <python:None>`
        :rtype: :class:`MarkerDefinition` or :obj:`None <python:None>`
        """
        return self._defs

    @defs.setter
    @class_sensitive(MarkerDefinition)
    def defs(self, value):
        self._defs = value

    @property
    def exporting(self) -> Optional[Exporting]:
        """Options to configure the export functionality enabled for the chart.

        :returns: The configuration of the chart's exporting functionality.
        :rtype: :class:`Exporting` or :obj:`None <python:None>`
        """
        return self._exporting

    @exporting.setter
    @class_sensitive(Exporting)
    def exporting(self, value):
        self._exporting = value

    @property
    def language(self) -> Optional[Language]:
        """Language object which can be used to configure the specific text to use in the
        chart.

        .. note::

          When working in JavaScript, the ``lang`` configuration is global and it can't be
          set on each chart initialization.

          Instead, use ``Highcharts.setOptions()`` to set it before any chart is
          initialized.

        :returns: A :class:`Language` object or :obj:`None <python:None>`
        :rtype: :class:`Language` or :obj:`None <python:None>`
        """
        return self._language

    @language.setter
    @class_sensitive(Language)
    def language(self, value):
        self._language = value

    @property
    def legend(self) -> Optional[Legend]:
        """The legend is a box containing a symbol and name for each series item or point
        item in the chart. Each series (or points in case of pie charts) is represented by
        a symbol and its name in the legend.

        .. seealso::

          It is possible to override the symbol creator function and create
          `custom legend symbols <https://jsfiddle.net/gh/get/library/pure/highcharts/highcharts/tree/master/samples/highcharts/studies/legend-custom-symbol/>`_.

        :returns: The :class:`Legend` configuration or :obj:`None <python:None>`
        :rtype: :class:`Legend` or :obj:`None <python:None>`
        """
        return self._legend

    @legend.setter
    @class_sensitive(Legend)
    def legend(self, value):
        self._legend = value

    @property
    def loading(self) -> Optional[Loading]:
        """The loading options control the appearance of the loading screen that covers
        the plot area on chart operations.

        This screen only appears after an explicit call to ``chart.showLoading()`` in the
        browser. It is a utility for developers to communicate to the end user that
        something is going on, for example while retrieving new data via an XHR
        connection.

        .. hint::

          The "Loading..." text itself is **not** part of this configuration object, but
          is instead part of the :meth:`.language <Options.language>` configuration.

        :returns: The configuration of the loading screen or :obj:`None <python:None>`
        :rtype: :class:`Loading` or :obj:`None <python:None>`
        """
        return self._loading

    @loading.setter
    @class_sensitive(Loading)
    def loading(self, value):
        self._loading = value

    @property
    def navigation(self) -> Optional[Navigation]:
        """A collection of options for buttons and menus appearing in the exporting
        module or in Stock Tools.

        :returns: The configuration of the navigation buttons.
        :rtype: ;class:`Navigation` or :obj:`None <python:None>`
        """
        return self._navigation

    @navigation.setter
    @class_sensitive(Navigation)
    def navigation(self, value):
        self._navigation = value

    @property
    def plot_options(self) -> Optional[PlotOptions]:
        """A wrapper object for configurations applied to each series type.

        The config objects for each series can also be overridden for each series item as
        given in the series array.

        Configuration options for the series are given in three levels:

          * Options for all series in a chart are given in the
            :meth:`series <PlotOptions.series>` property.
          * Options for all series of a specific type are given in the corresponding
            property for that type, for example
            :meth:`plot_options.line <PlotOptions.line>`.
          * Finally, options for one single series are given in the
            :meth:`series <Options.series>` array.

        :returns: Configurations for how series should be plotted / displayed.
        :rtype: :class:`PlotOptions` or :obj:`None <python:None>`
        """
        return self._plot_options

    @plot_options.setter
    @class_sensitive(PlotOptions)
    def plot_options(self, value):
        self._plot_options = value

    @property
    def responsive(self) -> Optional[Responsive]:
        """Rules to apply for different screen or chart sizes.

        .. note::

          Each rule specifies additional chart options.

        :returns: Rules to apply for different screen or chart sizes.
        :rtype: :class:`Responsive` or :obj:`None <python:None>`
        """
        return self._responsive

    @responsive.setter
    @class_sensitive(Responsive)
    def responsive(self, value):
        self._responsive = value

    @property
    def series(self) -> Optional[List[GenericTypeOptions]]:
        """Series options for specific data and the data itself.

        :returns: The series to display along with configuration and data.
        :rtype: :class:`Series` or :obj:`None <python:None>`
        """
        return self._series

    @series.setter
    def series(self, value):
        value = validators.iterable(value, allow_empty = True)
        default_series_type = None
        if self.chart:
            default_series_type = self.chart.type
        if not value:
            self._series = None
        else:
            self._series = [create_series_obj(x,
                                              default_type = default_series_type)
                            for x in value]

    @property
    def sonification(self) -> Optional[SonificationOptions]:
        """Configuration of global sonification settings for the entire chart.
        
        :rtype: :class:`SonificationOptions <highcharts_core.options.sonification.SonificationOptions>` or
          :obj:`None <python:None>`
        """
        return self._sonification
    
    @sonification.setter
    @class_sensitive(SonificationOptions)
    def sonification(self, value):
        self._sonification = value

    @property
    def subtitle(self) -> Optional[Subtitle]:
        """The chart's subtitle.

        .. note::

          This can be used both to display a subtitle below the main title, and to display
          random text anywhere in the chart.

        .. warning::

          The subtitle can be updated after chart initialization through the
          ``Chart.setTitle`` JavaScript method.

        :returns: Configuration of the chart's subtitle.
        :rtype: :class:`Subtitle` or :obj:`None <python:None>`
        """
        return self._subtitle

    @subtitle.setter
    @class_sensitive(Subtitle)
    def subtitle(self, value):
        self._subtitle = value

    @property
    def time(self) -> Optional[Time]:
        """Time options that can apply globally or to individual charts. These settings
        affect how datetime axes are laid out, how tooltips are formatted, how series
        :meth:`point_interval_unit <Series.point_interval_unit` works and how the
        Highcharts Stock range selector handles time.

        :returns: Configuration of applicable Time options.
        :rtype: :class:`Time` or :obj:`None <python:None>`
        """
        return self._time

    @time.setter
    @class_sensitive(Time)
    def time(self, value):
        self._time = value

    @property
    def title(self) -> Optional[Title]:
        """Options for configuring the chart's main title.

        :returns: Configuration of the chart's main title.
        :rtype: :class:`Title` or :obj:`None <python:None>`
        """
        return self._title

    @title.setter
    @class_sensitive(Title)
    def title(self, value):
        self._title = value

    @property
    def tooltip(self) -> Optional[Tooltip]:
        """Options for the tooltip that appears when the user hovers over a series or
        point.

        :returns: Configuration settings for tooltips to display above the chart.
        :rtype: :class:`Tooltip` or :obj:`None <python:None>`
        """
        return self._tooltip

    @tooltip.setter
    @class_sensitive(Tooltip)
    def tooltip(self, value):
        self._tooltip = value

    @property
    def x_axis(self) -> Optional[List[XAxis]]:
        """The X axis or category axis.

        Normally this is the horizontal axis, though if the chart is inverted this is the
        vertical axis.

        :returns: A collection of :class:`XAxis` objects
        :rtype: :class:`list <python:list>` of :class:`XAxis` or :obj:`None <python:None>`
        """
        return self._x_axis

    @x_axis.setter
    @class_sensitive(XAxis, force_iterable = True)
    def x_axis(self, value):
        self._x_axis = value

    @property
    def y_axis(self) -> Optional[List[YAxis]]:
        """The Y axis or value axis.

        Normally this is the vertical axis, though if the chart is inverted this is the
        horizontal axis.

        :returns: A collection of :class:`YAxis` objects
        :rtype: :class:`list <python:list>` of :class:`YAxis` or :obj:`None <python:None>`
        """
        return self._y_axis

    @y_axis.setter
    @class_sensitive(YAxis, force_iterable = True)
    def y_axis(self, value):
        self._y_axis = value

    def add_series(self, *series):
        """Adds ``series`` to the
        :meth:`series <highcharts_core.options.HighchartsOptions.series>` property.

        :param series: One or more :term:`series` instances (descended from
          :class:`SeriesBase <highcharts_core.options.series.base.SeriesBase>`) or an
          instance (e.g. :class:`dict <python:dict>`, :class:`str <python:str>`, etc.)
          coercable to one
        :type series: one or more
          :class:`SeriesBase <highcharts_core.options.series.base.SeriesBase>`
          or coercable

        """
        new_series = []
        for item in series:
            item_series = create_series_obj(item)
            new_series.append(item_series)

        if self.series:
            existing_series = [x for x in self.series]
        else:
            existing_series = []

        updated_series = existing_series + new_series

        self.series = updated_series

    @classmethod
    def from_series(cls, *series, kwargs = None):
        """Creates a new :class:`Options <highcharts_core.options.Options>` instance
        populated with ``series``.

        :param series: One or more :term:`series` instances (descended from
          :class:`SeriesBase <highcharts_core.options.series.base.SeriesBase>`) or an
          instance (e.g. :class:`dict <python:dict>`, :class:`str <python:str>`, etc.)
          coercable to one
        :type series: one or more
          :class:`SeriesBase <highcharts_core.options.series.base.SeriesBase>`
          or coercable

        :param kwargs: Other properties to use as keyword arguments for the instance to be
          created.

          .. warning::

            If ``kwargs`` sets the
            :meth:`.series <highcharts_core.options.Options.series>`
            property, that setting will be *overridden* by the contents of ``series``.

        :type kwargs: :class:`dict <python:dict>`

        :returns: A new :class:`Options <highcharts_core.options.Options>` instance
        :rtype: :class:`Options <highcharts_core.options.Options>`
        """
        kwargs = validators.dict(kwargs, allow_empty = True) or {}
        instance = cls(**kwargs)

        instance.add_series(series)


class HighchartsOptions(Options):
    """The Python representation of the `Highcharts <https://highcharts.com>`_
    ``options`` `configuration object <https://api.highcharts.com/highcharts/>`_."""

    def __init__(self, **kwargs):
        self._boost = None
        self._drilldown = None
        self._no_data = None
        self._pane = None
        self._z_axis = None

        self.boost = kwargs.get('boost', None)
        self.drilldown = kwargs.get('drilldown', None)
        self.no_data = kwargs.get('no_data', None)
        self.pane = kwargs.get('pane', None)
        self.z_axis = kwargs.get('z_axis', None)

        super().__init__(**kwargs)

    @property
    def boost(self) -> Optional[Boost]:
        """Options for the Boost module.

        The Boost module allows certain series types to be rendered by WebGL instead of
        the default SVG. This allows hundreds of thousands of data points to be rendered
        in milliseconds. In addition to the WebGL rendering it saves time by skipping
        processing and inspection of the data wherever possible.

        .. warning::

          This introduces some limitations to what features are available in boost mode.
          See
          `the docs <https://www.highcharts.com/docs/advanced-chart-features/boost-module>`_
          for details.

        .. note:

          In addition to the global boost option, each series has a ``boostThreshold``
          that defines when the boost should kick in.

        :returns: The :class:`Boost <Boost>` object.
        :rtype: :class:`Boost <highcharts.boost.Boost>`

        """
        return self._boost

    @boost.setter
    @class_sensitive(Boost)
    def boost(self, value):
        self._boost = value

    @property
    def drilldown(self) -> Optional[Drilldown]:
        """Options to configure :term:`drilldown` functionality in the chart, which
        enables users to inspect increasingly high resolution data by clicking on chart
        items like columns or pie slices.

        .. note::

          The drilldown feature requires the ``drilldown.js`` file to be loaded in the
          browser/client. This file is found in the modules directory of the download
          package, or online at
          `code.highcharts.com/modules/drilldown.js <code.highcharts.com/modules/drilldown.js>`_.

        :returns: The options to configure the chart's drill down functionality.
        :rtype: :class:`Drilldown` or :obj:`None <python:None>`
        """
        return self._drilldown

    @drilldown.setter
    @class_sensitive(Drilldown)
    def drilldown(self, value):
        self._drilldown = value

    @property
    def no_data(self) -> Optional[NoData]:
        """Options for displaying a message like "No data to display".

        .. warning::

          This feature requires the file ``no-data-to-display.js`` to be loaded in the
          page.

        .. tip::

          The actual text to display is set in the :meth:`language <Options.language>`
          options.

        :returns: Configuration of how to display a no data message.
        :rtype: :class:`NoData` or :obj:`None <python:None>`
        """
        return self._no_data

    @no_data.setter
    @class_sensitive(NoData)
    def no_data(self, value):
        self._no_data = value

    @property
    def pane(self) -> Optional[Pane]:
        """The pane serves as a container for axes and backgrounds for circular gauges and
        polar charts.

        :returns: The Pane configuration options.
        :rtype: :class:`Pane` or :obj:`None <python:None>`
        """
        return self._pane

    @pane.setter
    @class_sensitive(Pane)
    def pane(self, value):
        self._pane = value

    @property
    def z_axis(self) -> Optional[List[ZAxis]]:
        """The Z axis or depth axis for 3D plots.

        :returns: A collection of :class:`ZAxis` objects
        :rtype: :class:`list <python:list>` of :class:`ZAxis` or :obj:`None <python:None>`
        """
        return self._z_axis

    @z_axis.setter
    @class_sensitive(ZAxis, force_iterable = True)
    def z_axis(self, value):
        self._z_axis = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        as_dict = validators.dict(as_dict, allow_empty = True) or {}
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'annotations': as_dict.get('annotations', None),
            'boost': as_dict.get('boost', None),
            'caption': as_dict.get('caption', None),
            'chart': as_dict.get('chart', None),
            'color_axis': as_dict.get('colorAxis', None),
            'colors': as_dict.get('colors', None),
            'credits': as_dict.get('credits', None),
            'data': as_dict.get('data', None),
            'defs': as_dict.get('defs', None),
            'drilldown': as_dict.get('drilldown', None),
            'exporting': as_dict.get('exporting', None),
            'language': as_dict.get('lang', None) or as_dict.get('language', None),
            'legend': as_dict.get('legend', None),
            'loading': as_dict.get('loading', None),
            'navigation': as_dict.get('navigation', None),
            'no_data': as_dict.get('noData', None),
            'pane': as_dict.get('pane', None),
            'plot_options': as_dict.get('plotOptions', None),
            'responsive': as_dict.get('responsive', None),
            'series': as_dict.get('series', None),
            'sonification': as_dict.get('sonification', None),
            'subtitle': as_dict.get('subtitle', None),
            'time': as_dict.get('time', None),
            'title': as_dict.get('title', None),
            'tooltip': as_dict.get('tooltip', None),
            'x_axis': as_dict.get('xAxis', None),
            'y_axis': as_dict.get('yAxis', None),
            'z_axis': as_dict.get('zAxis', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'accessibility': self.accessibility,
            'annotations': self.annotations,
            'boost': self.boost,
            'caption': self.caption,
            'chart': self.chart,
            'colorAxis': self.color_axis,
            'colors': self.colors,
            'credits': self.credits,
            'data': self.data,
            'defs': self.defs,
            'drilldown': self.drilldown,
            'exporting': self.exporting,
            'lang': self.language,
            'legend': self.legend,
            'loading': self.loading,
            'navigation': self.navigation,
            'noData': self.no_data,
            'pane': self.pane,
            'plotOptions': self.plot_options,
            'responsive': self.responsive,
            'series': self.series,
            'sonification': self.sonification,
            'subtitle': self.subtitle,
            'time': self.time,
            'title': self.title,
            'tooltip': self.tooltip,
            'xAxis': self.x_axis,
            'yAxis': self.y_axis,
            'zAxis': self.z_axis
        }

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.utility_classes.javascript_functions import JavaScriptClass, \
  CallbackFunction


class Time(HighchartsMeta):
    """Time options that can apply globally or to individual charts. These settings
    affect how datetime axes are laid out, how tooltips are formatted, how series
    :meth:`point_interval_unit <Series.point_interval_unit` works and how the
    Highcharts Stock range selector handles time."""

    def __init__(self, **kwargs):
        self._Date = None
        self._get_timezone_offset = None
        self._moment = None
        self._timezone = None
        self._timezone_offset = None
        self._use_utc = None

        self.Date = kwargs.get('Date', None)
        self.get_timezone_offset = kwargs.get('get_timezone_offset', None)
        self.moment = kwargs.get('moment', None)
        self.timezone = kwargs.get('timezone', None)
        self.timezone_offset = kwargs.get('timezone_offset', None)
        self.use_utc = kwargs.get('use_utc', None)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'time'

    @property
    def Date(self) -> Optional[JavaScriptClass]:
        """A custom JavaScript ``Date`` class for more sophisticated date handling. For
        example, `JDate <https://github.com/tahajahangir/jdate>`_ can be used to handle
        Jalali dates. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`JavaScriptClass` or :obj:`None <python:None>`
        """
        return self._Date

    @Date.setter
    @class_sensitive(JavaScriptClass)
    def Date(self, value):
        self._Date = value

    @property
    def get_timezone_offset(self) -> Optional[CallbackFunction]:
        """A JavaScript callback function that returns the time zone offset for a given
        datetime. Defaults to :obj:`None <python:None>`.

        The function is expected to take a Unix timestamp (as number of milliseconds since
        January 1, 1970) as an argument and return the timezone offset in minutes.

        .. hint::

          This provides a hook for drawing time based charts in specific time zones using
          their local DST crossover dates, with the help of external libraries.

        .. warning::

          If :meth:`Time.timezone` is specified, then this custom function will be
          overridden by an automatic function leveraging the
          `moment.js <https://momentjs.com>`_ library.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._get_timezone_offset

    @get_timezone_offset.setter
    @class_sensitive(CallbackFunction)
    def get_timezone_offset(self, value):
        self._get_timezone_offset = value

    @property
    def moment(self) -> Optional[CallbackFunction]:
        """A JavaScript function that can be used to manually load the
        `moment.js <https://momentjs.com>`_ library when initializing a chart, rather than
        loading it from within ``window`` or a ``<script/>`` tag. Defaults to
        :obj:`None <python:None>`.

        .. note::

          The function expects no arguments and returns no result.

        .. warning::

          If loading ``moment.js`` is loaded from within a ``<script/>`` tag on the page,
          this setting can be left as :obj:`None <python:None>`. Highcharts will still be
          able to access and use the library.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._moment

    @moment.setter
    @class_sensitive(CallbackFunction)
    def moment(self, value):
        self._moment = value

    @property
    def timezone(self) -> Optional[str]:
        """Indicates the timezone in which to render the chart. Defaults to
        :obj:`None <python:None>`.

        If this value is provided, Highcharts will override the setting for
        :meth:`Time.get_timezone_offset` with a default function leveraging the
        `moment.js <https://momentjs.com>`_ library.

        .. warning::

          Requires `moment.js <https://momentjs.com>`_ to work properly. If ``moment.js``
          is not available, this will throw a Highcharts error in the JavaScript console
          but will *not* prevent the chart from rendering.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._timezone

    @timezone.setter
    def timezone(self, value):
        self._timezone = validators.string(value, allow_empty = True)

    @property
    def timezone_offset(self) -> Optional[int]:
        """The timezone offset to apply in minutes. Defaults to ``0``.

        .. hint::

          Use this to display UTC data in a predefined timezone.

        .. note::

          Positive values are west of UTC. Negative values are east of UTC, as in the
          ECMAScript
          `getTimezoneOffset() <https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/getTimezoneOffset>`_
          method.

        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._timezone_offset

    @timezone_offset.setter
    def timezone_offset(self, value):
        self._timezone_offset = validators.integer(value, allow_empty = True)

    @property
    def use_utc(self) -> Optional[bool]:
        """If ``True``, use UTC time for axis scaling, tickmark placement, and time
        display. Defaults to ``True``.

        .. hint::

          The advantage of using UTC is that the time displays equally regardless of the
          user agent's time zone settings. Local time can be used when the data is loaded
          in real time or when correct Daylight Saving Time transitions are required.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_utc

    @use_utc.setter
    def use_utc(self, value):
        if value is None:
            self._use_utc = None
        else:
            self._use_utc = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'Date': as_dict.get('Date', None),
            'get_timezone_offset': as_dict.get('getTimezoneOffset', None),
            'moment': as_dict.get('moment', None),
            'timezone': as_dict.get('timezone', None),
            'timezone_offset': as_dict.get('timezoneOffset', None),
            'use_utc': as_dict.get('useUTC', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'Date': self.Date,
            'getTimezoneOffset': self.get_timezone_offset,
            'moment': self.moment,
            'timezone': self.timezone,
            'timezoneOffset': self.timezone_offset,
            'useUTC': self.use_utc
        }

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import errors
from highcharts_core.decorators import class_sensitive, validate_types
from highcharts_core.metaclasses import HighchartsMeta, JavaScriptDict
from highcharts_core.utility_classes.animation import AnimationOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern
from highcharts_core.utility_classes.shadows import ShadowOptions
from highcharts_core.utility_classes.ast import TextPath
from highcharts_core.utility_classes.javascript_functions import CallbackFunction


class Filter(HighchartsMeta):
    """A declarative filter to control of which data labels to display.

    The declarative filter is designed for use when JavaScript callback functions are
    not available, like when the chart options require a pure JSON structure or for
    use with graphical editors. For programmatic control, use the
    :meth:`DataLabel.formatter` instead, and return ``undefined`` to disable a single
    data label."""

    def __init__(self, **kwargs):
        self._operator = None
        self._property = None
        self._value = None

        self.operator = kwargs.get('operator', None)
        self.property_ = kwargs.get('property_', None)
        self.value = kwargs.get('value', None)

    @property
    def operator(self) -> Optional[str]:
        """The operator to compare by. Defaults to :obj:`None <python:None>`.

        Accepts:

          * ``'>'``
          * ``'<'``
          * ``'>='``
          * ``'<='``
          * ``'=='``
          * ``'==='``

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._operator

    @operator.setter
    def operator(self, value):
        self._operator = validators.string(value, allow_empty = True)

    @property
    def property_(self) -> Optional[str]:
        """The point property to filter by. Defaults to :obj:`None <python:None>`.

        Point options are passed directly to properties, additionally there are ``y``
        value, ``percentage``.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._property

    @property_.setter
    def property_(self, value):
        self._property = validators.string(value, allow_empty = True)

    @property
    def value(self) -> Optional[int | float | Decimal]:
        """The value to compare against. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._value

    @value.setter
    def value(self, value_):
        self._value = validators.numeric(value_, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'operator': as_dict.get('operator', None),
            'property_': as_dict.get('property', None),
            'value': as_dict.get('value', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        return {
            'operator': self.operator,
            'property': self.property_,
            'value': self.value
        }


class DataLabel(HighchartsMeta):
    """Options for the series data labels, appearing next to each data point."""

    def __init__(self, **kwargs):
        self._align = None
        self._allow_overlap = None
        self._animation = None
        self._background_color = None
        self._border_color = None
        self._border_radius = None
        self._border_width = None
        self._class_name = None
        self._color = None
        self._crop = None
        self._defer = None
        self._enabled = None
        self._filter = None
        self._format = None
        self._formatter = None
        self._inside = None
        self._null_format = None
        self._null_formatter = None
        self._overflow = None
        self._padding = None
        self._position = None
        self._rotation = None
        self._shadow = None
        self._shape = None
        self._style = None
        self._text_path = None
        self._use_html = None
        self._vertical_align = None
        self._x = None
        self._y = None
        self._z = None

        self.align = kwargs.get('align', None)
        self.allow_overlap = kwargs.get('allow_overlap', None)
        self.animation = kwargs.get('animation', None)
        self.background_color = kwargs.get('background_color', None)
        self.border_color = kwargs.get('border_color', None)
        self.border_radius = kwargs.get('border_radius', None)
        self.border_width = kwargs.get('border_width', None)
        self.class_name = kwargs.get('class_name', None)
        self.color = kwargs.get('color', None)
        self.crop = kwargs.get('crop', None)
        self.defer = kwargs.get('defer', None)
        self.enabled = kwargs.get('enabled', None)
        self.filter = kwargs.get('filter', None)
        self.format = kwargs.get('format', None)
        self.formatter = kwargs.get('formatter', None)
        self.inside = kwargs.get('inside', None)
        self.null_format = kwargs.get('null_format', None)
        self.null_formatter = kwargs.get('null_formatter', None)
        self.overflow = kwargs.get('overflow', None)
        self.padding = kwargs.get('padding', None)
        self.position = kwargs.get('position', None)
        self.rotation = kwargs.get('rotation', None)
        self.shadow = kwargs.get('shadow', None)
        self.shape = kwargs.get('shape', None)
        self.style = kwargs.get('style', None)
        self.text_path = kwargs.get('text_path', None)
        self.use_html = kwargs.get('use_html', None)
        self.vertical_align = kwargs.get('vertical_align', None)
        self.x = kwargs.get('x', None)
        self.y = kwargs.get('y', None)
        self.z = kwargs.get('z', None)

    @property
    def align(self) -> Optional[str]:
        """The alignment of the data label compared to the point. Defaults to
        ``None``.

        Accepts:

          * ``'left'``
          * ``'center'``
          * ``'right'``

        .. hint::

          If right, the right side of the label should be touching the point.

        :returns: The alignment of the annotation's label.
        :rtype: :class:`str <python:str>`
        """
        return self._align

    @align.setter
    def align(self, value):
        if not value:
            self._align = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['left', 'center', 'right']:
                raise errors.HighchartsValueError(f'align must be either "left", '
                                                  f'"center", or "right". Was: {value}')

        self._align = value

    @property
    def allow_overlap(self) -> Optional[bool]:
        """If ``True``, data labels are allowed to overlap each other.

        Defaults to ``False``.

        .. hint::

          To make the labels less sensitive for overlapping, the :meth:`DataLabel.padding`
          can be set to ``0``.

        :returns: Flag indicating whether to allow data labels to overlap.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._allow_overlap

    @allow_overlap.setter
    def allow_overlap(self, value):
        if value is None:
            self._allow_overlap = None
        else:
            self._allow_overlap = bool(value)

    @property
    def animation(self) -> Optional[AnimationOptions]:
        """Enable or disable the initial animation for the data labels when a series is
        displayed.

        The animation can also be set as a configuration object. Please note that this
        option only applies to the initial animation of the series itself. For other
        animations, see :class:`Chart.animation` and the ``animation`` parameter under the
        (JavaScript) API methods. The following properties are supported:

          * ``defer``: The animation delay time in milliseconds.

        .. warning::

          Due to poor performance, animation is disabled in old IE browsers for several
          chart types.

        :rtype: :class:`AnimationOptions` or :obj:`None <python:None>`
        """
        return self._animation

    @animation.setter
    @class_sensitive(AnimationOptions)
    def animation(self, value):
        self._animation = value

    @property
    def background_color(self) -> Optional[str | Gradient | Pattern]:
        """The background color or gradient for the data label. Defaults to
        :obj:`None <python:None>`.

        :returns: The backgorund color for the data label.
        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern``, or
          :obj:`None <python:None>`
        """
        return self._background_color

    @background_color.setter
    def background_color(self, value):
        from highcharts_core import utility_functions
        self._background_color = utility_functions.validate_color(value)

    @property
    def border_color(self) -> Optional[str]:
        """The border color for the data label. Defaults to
        :obj:`None <python:None>`

        :returns: The border color for the data label.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._border_color

    @border_color.setter
    def border_color(self, value):
        self._border_color = validators.string(value, allow_empty = True)

    @property
    def border_radius(self) -> Optional[int | float | Decimal]:
        """The border radius (in pixels) applied to the data label. Defaults to
        ``0``.

        :returns: The border radius to apply to the data label.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_radius

    @border_radius.setter
    def border_radius(self, value):
        self._border_radius = validators.numeric(value, allow_empty = True)

    @property
    def border_width(self) -> Optional[int | float | Decimal]:
        """The border width (in pixels) applied to the data label. Defaults to
        ``0``.

        :returns: The border width to apply to the data label.
        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._border_width

    @border_width.setter
    def border_width(self, value):
        self._border_width = validators.numeric(value, allow_empty = True)

    @property
    def class_name(self) -> Optional[str]:
        """A classname to apply styling using CSS. Defaults to
        ``'highcharts-no-tooltip'``.

        :returns: The classname to apply to enable styling via CSS.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._class_name

    @class_name.setter
    def class_name(self, value):
        self._class_name = validators.string(value, allow_empty = True)

    @property
    def color(self) -> Optional[str]:
        """The text color for the data labels. Defaults to :obj:`None <python:None>`.

        .. note::

          For certain series types, like column or map, the data labels can be drawn
          inside the points. In this case the data label will be drawn with maximum
          contrast by default. Additionally, it will be given a ``text-outline`` style
          with the opposite color, to further increase the contrast. This can be
          overridden by setting the ``text-outline`` style to ``none`` in the
          :meth:`DataLabel.style` setting.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._color

    @color.setter
    def color(self, value):
        self._color = validators.string(value, allow_empty = True)

    @property
    def crop(self) -> Optional[bool]:
        """If ``True``, hide part of the data label that falls outside the plot
        area. Defaults to ``False``.

        .. note::

          By default, the data label is moved inside the plot area as per the
          :meth:`DataLabel.overflow` setting.

        :returns: Flag indicating whether to clip a data label that extends beyond
          the plot area.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._crop

    @crop.setter
    def crop(self, value):
        if value is None:
            self._crop = None
        else:
            self._crop = bool(value)

    @property
    def defer(self) -> Optional[bool | int]:
        """Whether to defer displaying the data labels until the initial series animation
        has finished. If :obj:`None <python:None>`, behaves as if set to ``True``.

        Setting to ``False`` renders the data label immediately.

        If set to ``True`` inherits the defer time set in
        :meth:`PlotOptions.series.animation`.

        If set to a number, defers the animation by that number of milliseconds.

        :rtype: :class:`bool <python:bool>` or :class:`int <python:int>` or
          :obj:`None <python:None>`
        """
        return self._defer

    @defer.setter
    def defer(self, value):
        if value is None:
            self._defer = None
        else:
            if value is True or value is False:
                self._defer = value
            else:
                self._defer = validators.integer(value)

    @property
    def enabled(self) -> Optional[bool]:
        """Enable or disable the data labels. Setting to :obj:`None <python:None>` behaves
        as if set to ``False``.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if value is None:
            self._enabled = None
        else:
            self._enabled = bool(value)

    @property
    def filter(self) -> Optional[Filter]:
        """A declarative filter to control of which data labels to display.

        The declarative filter is designed for use when JavaScript callback functions are
        not available, like when the chart options require a pure JSON structure or for
        use with graphical editors. For programmatic control, use the
        :meth:`DataLabel.formatter` instead, and return ``undefined`` to disable a single
        data label.

        :rtype: :class:`Filter` or :obj:`None <python:None>`
        """
        return self._filter

    @filter.setter
    @class_sensitive(Filter)
    def filter(self, value):
        self._filter = value

    @property
    def format(self) -> Optional[str]:
        """A format string to apply to the label. Defaults to
        ``'point.value'``.


        :returns: The format string to apply to the labels.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._format

    @format.setter
    def format(self, value):
        self._format = validators.string(value, allow_empty = True)

    @property
    def formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the data label. Defaults to
        :obj:`None <python:None>`.

        .. note::

          If a :meth:`DataLabel.format` is specified, the formatter will be ignored.

        :returns: A JavaScript callback function.
        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._formatter

    @formatter.setter
    @class_sensitive(CallbackFunction)
    def formatter(self, value):
        self._formatter = value

    @property
    def inside(self) -> Optional[bool]:
        """For points with an extent, like columns or map areas, whether to align the data
        label inside the box or to the actual value point. Defaults to
        ``:obj:`None <python:None>`, which behaves like ``False`` in most cases but
        ``True`` in stacked columns.

        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._inside

    @inside.setter
    def inside(self, value):
        if value is None:
            self._inside = None
        else:
            self._inside = bool(value)

    @property
    def null_format(self) -> Optional[str]:
        """Format for points with the value of ``null``. Defaults to
        :obj:`None <python:None>`.

        .. note::

          Works analogously to :meth:`DataLabel.format`.

        .. warning::

          Can only be applied only to series which support displaying null points.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._null_format

    @null_format.setter
    def null_format(self, value):
        self._null_format = validators.string(value, allow_empty = True)

    @property
    def null_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format the text of the data label for visible
        null points.

        .. note::

          Works analogously to :meth:`DataLabel.formatter`.

        .. warning::

          Can only be applied only to series which support displaying null points.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._null_formatter

    @null_formatter.setter
    @class_sensitive(CallbackFunction)
    def null_formatter(self, value):
        self._null_formatter = value

    @property
    def overflow(self) -> Optional[str]:
        """Configuration on how to handle a data label that overflows outside of
        the plot area.  Defaults to ``'justify'``,
        which aligns them inside the plot area. For columns and bars, this means the data
        label will be moved inside the bar.

        .. hint::

          To display data labels outside the plot area, set ``overflow`` to ``'allow'``
          and :meth:`DataLabel.crop` to ``False``.

        Accepts:

          * ``'justify'`` - which forces the label back into the plot area
          * ``'allow'`` - which allows data labels to overflow outside of the plot area

        .. note::

          The overflow treatment is also affected by the :meth:`DataLabel.crop`
          setting.

        :returns: Configuration of overflow setting.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._overflow

    @overflow.setter
    def overflow(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._overflow = None
        else:
            value = value.lower()
            if value not in ['justify', 'none']:
                raise errors.HighchartsValueError(f'overflow accepts "justify" or "none".'
                                                  f' Was: {value}')
            self._overflow = value

    @property
    def padding(self) -> Optional[int]:
        """The padding within the border box when either
        :meth:`DataLabel.border_width` or :meth:`DataLabel.background_color` is set.

        Defaults to ``5``.

        :returns: The padding to apply to the data label.
        :rtype: :class:`int <python:int>` or :obj:`None <python:None>`
        """
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = validators.numeric(value, allow_empty = True)

    @property
    def position(self) -> Optional[str]:
        """Aligns data labels relative to points. Defaults to
        ``'center'``.

        Accepts the following values:

          * ``'center'`` (the default)
          * ``'left'``
          * ``'right'``

        .. note::

          If ``center`` is not possible, aligns to ``right``.

        :rtype: :class:`str <python:str>`
        """
        return self._position

    @position.setter
    def position(self, value):
        if not value:
            self._position = None
        else:
            value = validators.string(value)
            value = value.lower()
            if value not in ['center', 'left', 'right']:
                raise errors.HighchartsValueError(f'position expects a value of "center",'
                                                  f' "left", or "right". Was: {value}')
            self._position = value

    @property
    def rotation(self) -> Optional[int | float | Decimal]:
        """Text rotation in degrees. Defaults to
        ``0``

        .. warning::

          Due to a more complex structure, backgrounds, borders and padding will be lost
          on a rotated data label.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        self._rotation = validators.numeric(value, allow_empty = True)

    @property
    def shadow(self) -> Optional[bool | ShadowOptions]:
        """Configuration for the shadow to apply to the data label box. Defaults to
        ``False``.

        If ``False``, no shadow is applied.

        :returns: The shadow configuration to apply or ``False``.
        :rtype: :class:`bool <python:bool>` or :class:`ShadowOptions` or
          :obj:`None <python:None>`
        """
        return self._shadow

    @shadow.setter
    def shadow(self, value):
        if value is None:
            self._shadow = None
        elif value is False:
            self._shadow = False
        else:
            value = validate_types(value,
                                   types = ShadowOptions)
            self._shadow = value

    @property
    def shape(self) -> Optional[str]:
        """The name of the symbol to use for the border around the label. Defaults to
        ``'square'``.

        Accepts:

          * ``'rect'``
          * ``'square'``
          * ``'circle'``
          * ``'diamond'``
          * ``'triangle'``
          * ``'callout'``

        :returns: The shape to use for the border around the label.
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._shape

    @shape.setter
    def shape(self, value):
        if not value:
            self._shape = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['callout',
                             'connector',
                             'rect',
                             'circle',
                             'diamond',
                             'triangle']:
                raise errors.HighchartsValueError(f'shape expects a supported annotation '
                                                  f'label shape. Was: {value}')
            self._shape = value

    @property
    def style(self) -> Optional[dict | str]:
        """CSS styling to apply to the annotation's label.

        The default color setting is ``"contrast"``, which is a pseudo color that
        Highcharts picks up and applies the maximum contrast to the underlying point item,
        for example the bar in a bar chart.

        ``textOutline`` is a pseudo property that applies an outline of the given width
        with the given color, which by default is the maximum contrast to the text. So a
        bright text color will result in a black text outline for maximum readability on
        a mixed background. In some cases, especially with grayscale text, the text
        outline doesn't work well, in which cases it can be disabled by setting it to
        ``"none"``. When :meth:`DataLabel.use_html` is ``True``, the ``textOutline`` will
        not be picked up. In this, case, the same effect can be acheived through the
        ``text-shadow`` CSS property.

        For some series types, where each point has an extent, like for example tree maps,
        the data label may overflow the point. There are two strategies for handling
        overflow. By default, the text will wrap to multiple lines. The other strategy is
        to set ``textOverflow`` to ellipsis, which will keep the text on one line plus it
        will break inside long words.

        :rtype: :class:`dict <python:dict>` or :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._style

    @style.setter
    def style(self, value):
        try:
            self._style = validators.dict(value, allow_empty = True)
        except (ValueError, TypeError):
            self._style = validators.string(value, allow_empty = True)

    @property
    def text_path(self) -> Optional[TextPath]:
        """Options for a label text which should follow marker's shape.

        .. note::

          Border and background are disabled for a label that follows a path.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._text_path

    @text_path.setter
    @class_sensitive(TextPath)
    def text_path(self, value):
        self._text_path = value

    @property
    def use_html(self) -> Optional[bool]:
        """If ``True``, will use HTML to render the data label. If ``False``, will
        use SVG or WebGL as applicable.

        Defaults to ``False``.

        :returns: Flag indicating whether to render data labels using HTML.
        :rtype: :class:`bool <python:bool>` or :obj:`None <python:None>`
        """
        return self._use_html

    @use_html.setter
    def use_html(self, value):
        if value is None:
            self._use_html = None
        else:
            self._use_html = bool(value)

    @property
    def vertical_align(self) -> Optional[str]:
        """The vertical alignment of the annotation's label. Defaults to
        :obj:`None <python:None>`.

        If :obj:`None <python:None>`, the alignment will depend on the data. For example,
        in a column chart, the label would be above positive values and below negative
        values.

        Accepts:

          * ``'bottom'``
          * ``'middle'``
          * ``'top'``
          * :obj:`None <python:None>`

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._vertical_align

    @vertical_align.setter
    def vertical_align(self, value):
        value = validators.string(value, allow_empty = True)
        if not value:
            self._vertical_align = None
        else:
            value = value.lower()
            if value not in ['bottom', 'middle', 'top']:
                raise errors.HighchartsValueError(f'vertical_align expects either "top", '
                                                  f'"middle", or "bottom". Was: {value}')
            self._vertical_align = value

    @property
    def x(self) -> Optional[int | float | Decimal]:
        """The x position offset of the label relative to the point. Defaults to
        ``0``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = validators.numeric(value, allow_empty = True)

    @property
    def y(self) -> Optional[int | float | Decimal]:
        """The y position offset of the label relative to the point. Defaults to
        ``None``.

        :rtype: numeric
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = validators.numeric(value, allow_empty = True)

    @property
    def z(self) -> Optional[int]:
        """The Z index of the data labels. Defaults to
        ``6``.

        If :obj:`None <python:None>`, will be placed above the series.

        .. hint::

          Use a Z index of ``2`` to display it behind the series.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._z

    @z.setter
    def z(self, value):
        self._z = validators.numeric(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'align': as_dict.get('align', None),
            'allow_overlap': as_dict.get('allowOverlap', None),
            'animation': as_dict.get('animation', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'crop': as_dict.get('crop', None),
            'defer': as_dict.get('defer', None),
            'enabled': as_dict.get('enabled', None),
            'filter': as_dict.get('filter', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'inside': as_dict.get('inside', None),
            'null_format': as_dict.get('nullFormat', None),
            'null_formatter': as_dict.get('nullFormatter', None),
            'overflow': as_dict.get('overflow', None),
            'padding': as_dict.get('padding', None),
            'position': as_dict.get('position', None),
            'rotation': as_dict.get('rotation', None),
            'shadow': as_dict.get('shadow', None),
            'shape': as_dict.get('shape', None),
            'style': as_dict.get('style', None),
            'text_path': as_dict.get('textPath', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
            'z': as_dict.get('z', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'align': self.align,
            'allowOverlap': self.allow_overlap,
            'animation': self.animation,
            'backgroundColor': self.background_color,
            'borderColor': self.border_color,
            'borderRadius': self.border_radius,
            'borderWidth': self.border_width,
            'className': self.class_name,
            'color': self.color,
            'crop': self.crop,
            'defer': self.defer,
            'enabled': self.enabled,
            'filter': self.filter,
            'format': self.format,
            'formatter': self.formatter,
            'inside': self.inside,
            'nullFormat': self.null_format,
            'nullFormatter': self.null_formatter,
            'overflow': self.overflow,
            'padding': self.padding,
            'position': self.position,
            'rotation': self.rotation,
            'shadow': self.shadow,
            'shape': self.shape,
            'style': self.style,
            'textPath': self.text_path,
            'useHTML': self.use_html,
            'verticalAlign': self.vertical_align,
            'x': self.x,
            'y': self.y,
            'z': self.z
        }

        return untrimmed


class SunburstDataLabel(DataLabel):
    """Variant of :class:`DataLabel` used for :term:`sunburst` series."""
    
    def __init__(self, **kwargs):
        self._rotation_mode = None
        
        self.rotation_mode = kwargs.get('rotation_mode', None)
        
        super().__init__(**kwargs)
        
    @property
    def rotation_mode(self) -> Optional[str]:
        """Determines how the data label will be rotated relative to the perimeter of the sunburst. 
        
        Valid values are:
        
          * ``'circular'``
          * ``'auto'``
          * ``'parallel'`` 
          * ``'perpendicular'``. 
          
        Defaults to ``'circular'``.
        
        .. note::

          When ``'circular'``, the best fit will be computed for the point, so that the label is curved around the 
          center when there is room for it, otherwise perpendicular. 
        
          The legacy ``'auto'`` option works similiarly to ``'circular'``, but instead of curving the labels, they are 
          tangented to the perimiter.
        
        .. warning::
        
          The :meth:`.rotation <highcharts_core.utility_classes.data_labels.SunburstDataLabel.rotation>` property 
          takes precedence over ``.rotation_mode``.
          
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._rotation_mode
    
    @rotation_mode.setter
    def rotation_mode(self, value):
        if not value:
            self._rotation_mode = None
        else:
            value = validators.string(value, allow_empty = False)
            value = value.lower()
            if value not in ['circular', 'auto', 'parallel', 'perpendicular']:
                raise errors.HighchartsValueError(f'if not empty, rotation_mode expects a value of either '
                                                  f'"circular", "auto", "parallel", or "perpendicular", '
                                                  f' but received "{str}".')

            self._rotation_mode = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'align': as_dict.get('align', None),
            'allow_overlap': as_dict.get('allowOverlap', None),
            'animation': as_dict.get('animation', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'crop': as_dict.get('crop', None),
            'defer': as_dict.get('defer', None),
            'enabled': as_dict.get('enabled', None),
            'filter': as_dict.get('filter', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'inside': as_dict.get('inside', None),
            'null_format': as_dict.get('nullFormat', None),
            'null_formatter': as_dict.get('nullFormatter', None),
            'overflow': as_dict.get('overflow', None),
            'padding': as_dict.get('padding', None),
            'position': as_dict.get('position', None),
            'rotation': as_dict.get('rotation', None),
            'shadow': as_dict.get('shadow', None),
            'shape': as_dict.get('shape', None),
            'style': as_dict.get('style', None),
            'text_path': as_dict.get('textPath', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
            'z': as_dict.get('z', None),
            
            'rotation_mode': as_dict.get('rotationMode', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'rotationMode': self.rotation_mode,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class OrganizationDataLabel(DataLabel):
    """Variant of :class:`DataLabel` used for :term:`organization` series."""
    
    def __init__(self, **kwargs):
        self._link_format = None
        self._link_formatter = None
        self._link_text_path = None
        
        self.link_format = kwargs.get('link_format', None)
        self.link_formatter = kwargs.get('link_formatter', None)
        self.link_text_path = kwargs.get('link_text_path', None)
        
        super().__init__(**kwargs)

    @property
    def link_format(self) -> Optional[str]:
        """The format string specifying what to show for links in the\rorganization chart.
        
        .. tip::
        
          Best to use with 
          :meth:`.link_text_path <highcharts_core.utility_classes.data_labels.OrganizationDataLabel.link_text_path>`
          enabled.
          
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._link_format
    
    @link_format.setter
    def link_format(self, value):
        self._link_format = validators.string(value, allow_empty = True)

    @property
    def link_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format data labels for links in the organization chart. 

        .. note::

          The :meth:`.link_format <highcharts_core.utility_classes.data_labels.OrganizationDataLabel.link_format>`
          property takes precedence over the ``link_formatter``.

        :rtype: :class:`CallbackFunction <highcharts_core.utility_classes.javascript_functions.CallbackFunction>` or
          :obj:`None <python:None>`
        """
        return self._link_formatter

    @link_formatter.setter
    @class_sensitive(CallbackFunction)
    def link_formatter(self, value):
        self._link_formatter = value

    @property
    def link_text_path(self) -> Optional[TextPath]:
        """Options for a label text which should follow the link's shape.

        .. note::

          Border and background are disabled for a label that follows a path.

        :rtype: :class:`TextPath <highcharts_core.utility_classes.ast.TextPath>` or :obj:`None <python:None>`
        """
        return self._link_text_path

    @link_text_path.setter
    @class_sensitive(TextPath)
    def link_text_path(self, value):
        self._link_text_path = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'align': as_dict.get('align', None),
            'allow_overlap': as_dict.get('allowOverlap', None),
            'animation': as_dict.get('animation', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'crop': as_dict.get('crop', None),
            'defer': as_dict.get('defer', None),
            'enabled': as_dict.get('enabled', None),
            'filter': as_dict.get('filter', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'inside': as_dict.get('inside', None),
            'null_format': as_dict.get('nullFormat', None),
            'null_formatter': as_dict.get('nullFormatter', None),
            'overflow': as_dict.get('overflow', None),
            'padding': as_dict.get('padding', None),
            'position': as_dict.get('position', None),
            'rotation': as_dict.get('rotation', None),
            'shadow': as_dict.get('shadow', None),
            'shape': as_dict.get('shape', None),
            'style': as_dict.get('style', None),
            'text_path': as_dict.get('textPath', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
            'z': as_dict.get('z', None),
            
            'link_format': as_dict.get('linkFormat', None),
            'link_formatter': as_dict.get('linkFormatter', None),
            'link_text_path': as_dict.get('linkTextPath', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'linkFormat': self.link_format,
            'linkFormatter': self.link_formatter,
            'linkTextPath': self.link_text_path,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class NodeDataLabel(DataLabel):
    """Variant of :class:`DataLabel` used for node-based charts/diagrams."""

    def __init__(self, **kwargs):
        self._node_format = None
        self._node_formatter = None

        self.node_format = kwargs.get('node_format', None)
        self.node_formatter = kwargs.get('node_formatter', None)

        super().__init__(**kwargs)

    @property
    def node_format(self) -> Optional[str]:
        """The format string which determines what to render for nodes in a sankey,
        organization, or similar diagram. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._node_format

    @node_format.setter
    def node_format(self, value):
        self._node_format = validators.string(value, allow_empty = True)

    @property
    def node_formatter(self) -> Optional[CallbackFunction]:
        """JavaScript callback function to format data labels for nodes in a sankey or
        organization diagram. Defaults to :obj:`None <python:None>`.

        .. note::

          The :meth:`node_format <NodeDataLabel.node_format>` takes precedence over the
          ``node_formatter``.

        :rtype: :class:`CallbackFunction` or :obj:`None <python:None>`
        """
        return self._node_formatter

    @node_formatter.setter
    @class_sensitive(CallbackFunction)
    def node_formatter(self, value):
        self._node_formatter = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'align': as_dict.get('align', None),
            'allow_overlap': as_dict.get('allowOverlap', None),
            'animation': as_dict.get('animation', None),
            'background_color': as_dict.get('backgroundColor', None),
            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'class_name': as_dict.get('className', None),
            'color': as_dict.get('color', None),
            'crop': as_dict.get('crop', None),
            'defer': as_dict.get('defer', None),
            'enabled': as_dict.get('enabled', None),
            'filter': as_dict.get('filter', None),
            'format': as_dict.get('format', None),
            'formatter': as_dict.get('formatter', None),
            'inside': as_dict.get('inside', None),
            'null_format': as_dict.get('nullFormat', None),
            'null_formatter': as_dict.get('nullFormatter', None),
            'overflow': as_dict.get('overflow', None),
            'padding': as_dict.get('padding', None),
            'position': as_dict.get('position', None),
            'rotation': as_dict.get('rotation', None),
            'shadow': as_dict.get('shadow', None),
            'shape': as_dict.get('shape', None),
            'style': as_dict.get('style', None),
            'text_path': as_dict.get('textPath', None),
            'use_html': as_dict.get('useHTML', None),
            'vertical_align': as_dict.get('verticalAlign', None),
            'x': as_dict.get('x', None),
            'y': as_dict.get('y', None),
            'z': as_dict.get('z', None),

            'node_format': as_dict.get('nodeFormat', None),
            'node_formatter': as_dict.get('nodeFormatter', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'nodeFormat': self.node_format,
            'nodeFormatter': self.node_formatter,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls) or {}
        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional

from validator_collection import validators

from highcharts_core.decorators import class_sensitive
from highcharts_core.metaclasses import HighchartsMeta
from highcharts_core.options.annotations import Annotation
from highcharts_core.options.navigation.bindings import Bindings
from highcharts_core.utility_classes.breadcrumbs import BreadcrumbOptions
from highcharts_core.utility_classes.buttons import ButtonConfiguration
from highcharts_core.utility_classes.events import NavigationEvents


class NavigationBase(HighchartsMeta):
    """A collection of options for buttons and menus appearing in the exporting
    module or in Stock Tools.
    """

    def __init__(self, **kwargs):
        self._annotation_options = None
        self._bindings = None
        self._bindings_class_name = None
        self._button_options = None
        self._events = None
        self._icons_url = None

        self.annotation_options = kwargs.get('annotation_options', None)
        self.bindings = kwargs.get('bindings', None)
        self.bindings_class_name = kwargs.get('bindings_class_name', None)
        self.button_options = kwargs.get('button_options', None)
        self.events = kwargs.get('events', None)
        self.icons_url = kwargs.get('icons_url', None)

        super().__init__(**kwargs)

    @property
    def annotation_options(self) -> Optional[Annotation]:
        """Additional options to be applied to all annotations.

        :rtype: :class:`NavigationAnnotationOptions` or :obj:`None <python:None>`
        """
        return self._annotation_options

    @annotation_options.setter
    @class_sensitive(Annotation)
    def annotation_options(self, value):
        self._annotation_options = value

    @property
    def bindings(self) -> Optional[Bindings]:
        """JavaScript event bindings for custom HTML buttons.

        :rtype: :class:`Bindings` or :obj:`None <python:None>`
        """
        return self._bindings

    @bindings.setter
    @class_sensitive(Bindings)
    def bindings(self, value):
        self._bindings = value

    @property
    def bindings_class_name(self) -> Optional[str]:
        """A CSS class name where all bindings will be attached to. Defaults to
        ``'highcharts-bindings-container'``.

        .. hint::

          Multiple charts on the same page should have separate class names to prevent
          duplicating events.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._bindings_class_name

    @bindings_class_name.setter
    def bindings_class_name(self, value):
        self._bindings_class_name = validators.string(value, allow_empty = True)

    @property
    def button_options(self) -> Optional[ButtonConfiguration]:
        """Configuration options for navigation buttons.

        :rtype: :class:`ButtonOptions`
        """
        return self._button_options

    @button_options.setter
    @class_sensitive(ButtonConfiguration)
    def button_options(self, value):
        self._button_options = value

    @property
    def events(self) -> Optional[NavigationEvents]:
        """Events to communicate between Stock Tools and custom GUI.

        :rtype: :class:`NavigationEvents` or :obj:`None <python:None>`
        """
        return self._events

    @events.setter
    @class_sensitive(NavigationEvents)
    def events(self, value):
        self._events = value

    @property
    def icons_url(self) -> Optional[str]:
        """Path where Highcharts will look for icons. Defaults to
        ``'https://code.highcharts.com/@product.version@/gfx/stock-icons/'``.

        Change this to use icons from a different server.

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._icons_url

    @icons_url.setter
    def icons_url(self, value):
        self._icons_url = validators.string(value, allow_empty = True)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'annotation_options': as_dict.get('annotationOptions', None),
            'bindings': as_dict.get('bindings', None),
            'bindings_class_name': as_dict.get('bindingsClassName', None),
            'button_options': as_dict.get('buttonOptions', None),
            'events': as_dict.get('events', None),
            'icons_url': as_dict.get('iconsURL', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'annotationOptions': self.annotation_options,
            'bindings': self.bindings,
            'bindingsClassName': self.bindings_class_name,
            'buttonOptions': self.button_options,
            'events': self.events,
            'iconsURL': self.icons_url
        }

        return untrimmed


class Navigation(NavigationBase):
    """A collection of options for buttons and menus appearing in the exporting
    module or in Stock Tools.
    """

    def __init__(self, **kwargs):
        self._breadcrumbs = None

        self.breadcrumbs = kwargs.get('breadcrumbs', None)

        super().__init__(**kwargs)

    @property
    def _dot_path(self) -> Optional[str]:
        """The dot-notation path to the options key for the current class.
        
        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return 'navigation'

    @property
    def breadcrumbs(self) -> Optional[BreadcrumbOptions]:
        """Options for breadcrumbs.

        .. note::

          Breadcrumbs general options are defined in :meth:`Navigation.breadcrumbs`.

          Specific options for drilldown are set in `:meth:Drilldown.breadcrumbs` and for
          tree-like series traversing, in
          :meth:`PlotOptions[series].breadbrumbs <Series.breadcrumbs>`.

        :rtype: :class:`BreadcrumbOptions` or :obj:`None <python:None>`
        """
        return self._breadcrumbs

    @breadcrumbs.setter
    @class_sensitive(BreadcrumbOptions)
    def breadcrumbs(self, value):
        self._breadcrumbs = value

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'annotation_options': as_dict.get('annotationOptions', None),
            'bindings': as_dict.get('bindings', None),
            'bindings_class_name': as_dict.get('bindingsClassName', None),
            'button_options': as_dict.get('buttonOptions', None),
            'events': as_dict.get('events', None),
            'icons_url': as_dict.get('iconsURL', None),

            'breadcrumbs': as_dict.get('breadcrumbs', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'breadcrumbs': self.breadcrumbs,
        }

        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants
from highcharts_core.options.plot_options.series import SeriesOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class AreaOptions(SeriesOptions):
    """General options to apply to all Area series types.

    .. figure:: ../../../_static/area-example.png
      :alt: Area Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._fill_color = None
        self._fill_opacity = None
        self._line_color = None
        self._negative_fill_color = None
        self._track_by_area = None

        self.fill_color = kwargs.get('fill_color', None)
        self.fill_opacity = kwargs.get('fill_opacity', None)
        self.line_color = kwargs.get('line_color', None)
        self.negative_fill_color = kwargs.get('negative_fill_color', None)
        self.track_by_area = kwargs.get('track_by_area', None)

        super().__init__(**kwargs)

    @property
    def fill_color(self) -> Optional[str | Gradient | Pattern | constants.EnforcedNullType]:
        """Fill color or gradient for the area. When :class:`EnforcedNullType`, the
        series' color is used with the series'
        :meth:`fill_opacity <AreaOptions.fill_opacity>`.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`EnforcedNullType`
        """
        return self._fill_color

    @fill_color.setter
    def fill_color(self, value):
        from highcharts_core import utility_functions
        self._fill_color = utility_functions.validate_color(value)

    @property
    def fill_opacity(self) -> Optional[int | float | Decimal]:
        """Fill opacity for the area. Defaults to ``0.75``.

        When you set an explicit :meth:`fill_color <AreaOptions.fill_color>`, the
        ``fill_opacity`` is not applied. Instead, you should define the opacity in the
        :meth:`fill_color <AreaOptions.fill_color>` with an rgba color definition.

        The ``fill_opacity`` setting, also the default setting, overrides the alpha
        component of the color setting.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._fill_opacity

    @fill_opacity.setter
    def fill_opacity(self, value):
        self._fill_opacity = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def line_color(self) -> Optional[str | Gradient | Pattern]:
        """A separate color for the graph line. When :obj:`None <python:None>`, by default
        the line takes the color of the series, but the ``line_color`` setting allows
        setting a separate color for the line without altering the
        :meth:`fill_color <AreaOptions.fill_color>`.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._line_color

    @line_color.setter
    def line_color(self, value):
        from highcharts_core import utility_functions
        self._line_color = utility_functions.validate_color(value)

    @property
    def negative_fill_color(self) -> Optional[str | Gradient | Pattern]:
        """A separate color for the negative part of the area.

        :rtype: :obj:`None <python:None>`, :class:`Gradient`, :class:`Pattern`, or
          :class:`str <python:str>`
        """
        return self._negative_fill_color

    @negative_fill_color.setter
    def negative_fill_color(self, value):
        from highcharts_core import utility_functions
        self._negative_fill_color = utility_functions.validate_color(value)

    @property
    def track_by_area(self) -> Optional[bool]:
        """When ``True``, the whole area should respond to mouseover tooltips and other
        mouse or touch events. When ``False``, only the line responds to mouse/touch
        events. Defaults to ``False``.

        :rtype: :class:`bool <python:bool>`
        """
        return self._track_by_area

    @track_by_area.setter
    def track_by_area(self, value):
        if value is None:
            self._track_by_area = None
        else:
            self._track_by_area = bool(value)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'fill_color': as_dict.get('fillColor', None),
            'fill_opacity': as_dict.get('fillOpacity', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'negative_fill_color': as_dict.get('negativeFillColor', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'track_by_area': as_dict.get('trackByArea', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'fillColor': self.fill_color,
            'fillOpacity': self.fill_opacity,
            'lineColor': self.line_color,
            'negativeFillColor': self.negative_fill_color,
            'trackByArea': self.track_by_area,
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class AreaRangeOptions(AreaOptions):
    """General options to apply to all AreaRange series types. The area range series
    is a carteseian series with higher and lower values for each point along an X
    axis, where the area between the values is shaded.

    .. figure:: ../../../_static/arearange-example.png
      :alt: AreaRange Example Chart
      :align: center

    """
    pass


class AreaSplineOptions(AreaOptions):
    """General options to apply to all AreaSpline series types. The area spline series
    is an area series where the graph between the points is smoothed into a spline.

    .. figure:: ../../../_static/areaspline-example.png
      :alt: AreaSpline Example Chart
      :align: center

    """
    pass


class AreaSplineRangeOptions(AreaOptions):
    """General options to apply to all AreaSplineRange series types. The area spline
    range series is a carteseian series type with higher and lower Y values along an X
    axis. The area inside the range is colored, and the graph outlining the area is a
    smoothed spline."""
    pass


class LineOptions(AreaOptions):
    """General options to apply to all Line series types.

    A line series displays information as a series of data points connected by
    straight line segments.

    .. figure:: ../../../_static/line-example.png
      :alt: Line Example Chart
      :align: center

    """
    pass


class StreamGraphOptions(AreaOptions):
    """General options to apply to all Stream Graph series types.

    A streamgraph is a type of stacked area graph which is displaced around a central
    axis, resulting in a flowing, organic shape.

    .. figure:: ../../../_static/streamgraph-example.png
      :alt: StreamGraph Example Chart
      :align: center

    """
    pass



from typing import Optional
from decimal import Decimal

from validator_collection import validators

from highcharts_core import constants, errors
from highcharts_core.decorators import class_sensitive
from highcharts_core.options.plot_options.bar import BarOptions
from highcharts_core.options.plot_options.drag_drop import BoxPlotDragDropOptions
from highcharts_core.utility_classes.gradients import Gradient
from highcharts_core.utility_classes.patterns import Pattern


class BoxPlotOptions(BarOptions):
    """General options to apply to all Box Plot series types.

    A box plot is a convenient way of depicting groups of data through their
    five-number summaries:

      * the smallest observation (sample minimum),
      * lower quartile (Q1),
      * median (Q2),
      * upper quartile (Q3), and
      * largest observation (sample maximum).

    .. figure:: ../../../_static/boxplot-example.png
      :alt: Box Plot Example Chart
      :align: center

    """

    def __init__(self, **kwargs):
        self._box_dash_style = None
        self._median_color = None
        self._median_dash_style = None
        self._median_width = None
        self._stem_dash_style = None
        self._stem_width = None
        self._whisker_color = None
        self._whisker_dash_style = None
        self._whisker_length = None
        self._whisker_width = None

        self.box_dash_style = kwargs.get('box_dash_style', None)
        self.median_color = kwargs.get('median_color', None)
        self.median_dash_style = kwargs.get('median_dash_style', None)
        self.median_width = kwargs.get('median_width', None)
        self.stem_dash_style = kwargs.get('stem_dash_style', None)
        self.stem_width = kwargs.get('stem_width', None)
        self.whisker_color = kwargs.get('whisker_color', None)
        self.whisker_dash_style = kwargs.get('whisker_dash_style', None)
        self.whisker_length = kwargs.get('whisker_length', None)
        self.whisker_width = kwargs.get('whisker_width', None)

        super().__init__(**kwargs)

    @property
    def box_dash_style(self) -> Optional[str]:
        """The dash style of the box.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._box_dash_style

    @box_dash_style.setter
    def box_dash_style(self, value):
        if not value:
            self._box_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'box_dash_style expects a recognized '
                                                  f'value, but received: {value}')
            self._box_dash_style = value

    @property
    def drag_drop(self) -> Optional[BoxPlotDragDropOptions]:
        """The draggable-points module allows points to be moved around or modified in the
        chart.

        In addition to the options mentioned under the dragDrop API structure, the module
        fires three (JavaScript) events:

          * ``point.dragStart``
          * ``point.drag``
          * ``point.drop``

        :rtype: :class:`DragDropOptions` or :obj:`None <python:None>`
        """
        return self._drag_drop

    @drag_drop.setter
    @class_sensitive(BoxPlotDragDropOptions)
    def drag_drop(self, value):
        self._drag_drop = value

    @property
    def median_color(self) -> Optional[str | Gradient]:
        """The color of the median line. If :obj:`None <python:None>`, the general series
        color applies. Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, or :obj:`None <python:None>`
        """
        return self._median_color

    @median_color.setter
    def median_color(self, value):
        from highcharts_core import utility_functions
        self._median_color = utility_functions.validate_color(value)

    @property
    def median_dash_style(self) -> Optional[str]:
        """The dash style of the median. Defaults to ``'Solid'``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._median_dash_style

    @median_dash_style.setter
    def median_dash_style(self, value):
        if not value:
            self._median_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'median_dash_style expects a '
                                                  f' recognized value, but received: '
                                                  f'{value}')
            self._median_dash_style = value

    @property
    def median_width(self) -> Optional[int | float | Decimal]:
        """The pixel width of the median line. If :obj:`None <python:None>`, the
        :meth:`line_width <BoxPlotOptions.line_width>` is used. Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._median_width

    @median_width.setter
    def median_width(self, value):
        self._median_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def stem_dash_style(self) -> Optional[str]:
        """The dash style of the :term:`stem`, the vertical line extending from the box to
        the whiskers. Defaults to ``'Solid'``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._stem_dash_style

    @stem_dash_style.setter
    def stem_dash_style(self, value):
        if not value:
            self._stem_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'stem_dash_style expects a recognized'
                                                  f' value, but received: {value}')
            self._stem_dash_style = value

    @property
    def stem_width(self) -> Optional[int | float | Decimal]:
        """The pixel width of the stem line. If :obj:`None <python:None>`, the
        :meth:`line_width <BoxPlotOptions.line_width>` is used. Defaults to
        :obj:`None <python:None>`.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._stem_width

    @stem_width.setter
    def stem_width(self, value):
        self._stem_width = validators.numeric(value,
                                                allow_empty = True,
                                                minimum = 0)

    @property
    def whisker_color(self) -> Optional[str | Gradient | Pattern]:
        """The color of the :term:`whiskers <whisker>`, the horizontal lines marking low
        and high values. When :obj:`None <python:None>`, the general series color is used.
        Defaults to :obj:`None <python:None>`.

        :rtype: :class:`str <python:str>`, :class:`Gradient`, :class:`Pattern`, or
          :obj:`None <python:None>`
        """
        return self._whisker_color

    @whisker_color.setter
    def whisker_color(self, value):
        from highcharts_core import utility_functions
        self._whisker_color = utility_functions.validate_color(value)

    @property
    def whisker_dash_style(self) -> Optional[str]:
        """The dash style of the whiskers. Defaults to ``'Solid'``.

        Accepts one of the following values:

          * 'Dash',
          * 'DashDot',
          * 'Dot',
          * 'LongDash',
          * 'LongDashDot',
          * 'LongDashDotDot',
          * 'ShortDash',
          * 'ShortDashDot',
          * 'ShortDashDotDot',
          * 'ShortDot',
          * 'Solid'

        :rtype: :class:`str <python:str>` or :obj:`None <python:None>`
        """
        return self._whisker_dash_style

    @whisker_dash_style.setter
    def whisker_dash_style(self, value):
        if not value:
            self._whisker_dash_style = None
        else:
            value = validators.string(value)
            if value not in constants.SUPPORTED_DASH_STYLE_VALUES:
                raise errors.HighchartsValueError(f'whisker_dash_style expects a '
                                                  f'recognized value, but received: '
                                                  f'{value}')
            self._whisker_dash_style = value

    @property
    def whisker_length(self) -> Optional[str | int | float | Decimal]:
        """The length of the whiskers, the horizontal lines marking low and high values.
        It can be a numerical pixel value, or a percentage value of the box width. Set to
        ``0`` to disable whiskers entirely.

        Defaults to ``'50%'``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._whisker_length

    @whisker_length.setter
    def whisker_length(self, value):
        if value is None:
            self._whisker_length = None
        else:
            try:
                value = validators.string(value)
                if '%' not in value:
                    raise errors.HighchartsValueError('whisker_length expects either a '
                                                      'number of pixels or a percentage '
                                                      'string. No "%" character was '
                                                      'found.')
            except TypeError:
                value = validators.numeric(value,
                                           allow_empty = True,
                                           minimum = 0)
            self._whisker_length = value

    @property
    def whisker_width(self) -> Optional[int | float | Decimal]:
        """The pixel width of the whisker line. If :obj:`None <python:None>`, the
        :meth:`line_width <BoxPlotOptions.line_width>` is used. Defaults to ``2``.

        :rtype: numeric or :obj:`None <python:None>`
        """
        return self._whisker_width

    @whisker_width.setter
    def whisker_width(self, value):
        self._whisker_width = validators.numeric(value,
                                                 allow_empty = True,
                                                 minimum = 0)

    @classmethod
    def _get_kwargs_from_dict(cls, as_dict):
        kwargs = {
            'accessibility': as_dict.get('accessibility', None),
            'allow_point_select': as_dict.get('allowPointSelect', None),
            'animation': as_dict.get('animation', None),
            'class_name': as_dict.get('className', None),
            'clip': as_dict.get('clip', None),
            'color': as_dict.get('color', None),
            'cursor': as_dict.get('cursor', None),
            'custom': as_dict.get('custom', None),
            'dash_style': as_dict.get('dashStyle', None),
            'data_labels': as_dict.get('dataLabels', None),
            'description': as_dict.get('description', None),
            'enable_mouse_tracking': as_dict.get('enableMouseTracking', None),
            'events': as_dict.get('events', None),
            'include_in_data_export': as_dict.get('includeInDataExport', None),
            'keys': as_dict.get('keys', None),
            'label': as_dict.get('label', None),
            'legend_symbol': as_dict.get('legendSymbol', None),
            'linked_to': as_dict.get('linkedTo', None),
            'marker': as_dict.get('marker', None),
            'on_point': as_dict.get('onPoint', None),
            'opacity': as_dict.get('opacity', None),
            'point': as_dict.get('point', None),
            'point_description_formatter': as_dict.get('pointDescriptionFormatter', None),
            'selected': as_dict.get('selected', None),
            'show_checkbox': as_dict.get('showCheckbox', None),
            'show_in_legend': as_dict.get('showInLegend', None),
            'skip_keyboard_navigation': as_dict.get('skipKeyboardNavigation', None),
            'sonification': as_dict.get('sonification', None),
            'states': as_dict.get('states', None),
            'sticky_tracking': as_dict.get('stickyTracking', None),
            'threshold': as_dict.get('threshold', None),
            'tooltip': as_dict.get('tooltip', None),
            'turbo_threshold': as_dict.get('turboThreshold', None),
            'visible': as_dict.get('visible', None),

            'animation_limit': as_dict.get('animationLimit', None),
            'boost_blending': as_dict.get('boostBlending', None),
            'boost_threshold': as_dict.get('boostThreshold', None),
            'color_axis': as_dict.get('colorAxis', None),
            'color_index': as_dict.get('colorIndex', None),
            'color_key': as_dict.get('colorKey', None),
            'connect_ends': as_dict.get('connectEnds', None),
            'connect_nulls': as_dict.get('connectNulls', None),
            'crisp': as_dict.get('crisp', None),
            'crop_threshold': as_dict.get('cropThreshold', None),
            'data_sorting': as_dict.get('dataSorting', None),
            'drag_drop': as_dict.get('dragDrop', None),
            'fill_color': as_dict.get('fillColor', None),
            'fill_opacity': as_dict.get('fillOpacity', None),
            'find_nearest_point_by': as_dict.get('findNearestPointBy', None),
            'get_extremes_from_all': as_dict.get('getExtremesFromAll', None),
            'linecap': as_dict.get('linecap', None),
            'line_color': as_dict.get('lineColor', None),
            'line_width': as_dict.get('lineWidth', None),
            'negative_color': as_dict.get('negativeColor', None),
            'negative_fill_color': as_dict.get('negativeFillColor', None),
            'point_interval': as_dict.get('pointInterval', None),
            'point_interval_unit': as_dict.get('pointIntervalUnit', None),
            'point_placement': as_dict.get('pointPlacement', None),
            'point_start': as_dict.get('pointStart', None),
            'relative_x_value': as_dict.get('relativeXValue', None),
            'shadow': as_dict.get('shadow', None),
            'soft_threshold': as_dict.get('softThreshold', None),
            'stacking': as_dict.get('stacking', None),
            'step': as_dict.get('step', None),
            'track_by_area': as_dict.get('trackByArea', None),
            'zone_axis': as_dict.get('zoneAxis', None),
            'zones': as_dict.get('zones', None),

            'border_color': as_dict.get('borderColor', None),
            'border_radius': as_dict.get('borderRadius', None),
            'border_width': as_dict.get('borderWidth', None),
            'center_in_category': as_dict.get('centerInCategory', None),
            'color_by_point': as_dict.get('colorByPoint', None),
            'colors': as_dict.get('colors', None),
            'grouping': as_dict.get('grouping', None),
            'group_padding': as_dict.get('groupPadding', None),
            'max_point_width': as_dict.get('maxPointWidth', None),
            'min_point_length': as_dict.get('minPointLength', None),
            'point_padding': as_dict.get('pointPadding', None),
            'point_range': as_dict.get('pointRange', None),
            'point_width': as_dict.get('pointWidth', None),

            'depth': as_dict.get('depth', None),
            'edge_color': as_dict.get('edgeColor', None),
            'edge_width': as_dict.get('edgeWidth', None),
            'group_z_padding': as_dict.get('groupZPadding', None),

            'box_dash_style': as_dict.get('boxDashStyle', None),
            'median_color': as_dict.get('medianColor', None),
            'median_dash_style': as_dict.get('medianDashStyle', None),
            'median_width': as_dict.get('medianWidth', None),
            'stem_dash_style': as_dict.get('stemDashStyle', None),
            'stem_width': as_dict.get('stemWidth', None),
            'whisker_color': as_dict.get('whiskerColor', None),
            'whisker_dash_style': as_dict.get('whiskerDashStyle', None),
            'whisker_length': as_dict.get('whiskerLength', None),
            'whisker_width': as_dict.get('whiskerWidth', None)
        }

        return kwargs

    def _to_untrimmed_dict(self, in_cls = None) -> dict:
        untrimmed = {
            'boxDashStyle': self.box_dash_style,
            'medianColor': self.median_color,
            'medianDashStyle': self.median_dash_style,
            'medianWidth': self.median_width,
            'stemDashStyle': self.stem_dash_style,
            'stemWidth': self.stem_width,
            'whiskerColor': self.whisker_color,
            'whiskerDashStyle': self.whisker_dash_style,
            'whiskerLength': self.whisker_length,
            'whiskerWidth': self.whisker_width
        }
        parent_as_dict = super()._to_untrimmed_dict(in_cls = in_cls)

        for key in parent_as_dict:
            untrimmed[key] = parent_as_dict[key]

        return untrimmed


class ErrorBarOptions(BoxPlotOptions):
    """General options to apply to all Error Bar series types.

    Error bars are a graphical representation of the variability of data and are used
    on graphs to indicate the error, or uncertainty in a reported measurement.

    .. figure:: ../../../_static/errorbar-example.png
      :alt: Error Bar Example Chart
      :align: center

    """
    pass
